import copy
from typing import Any
from typing import Dict
from typing import List


async def update_key_tags(
    hub,
    ctx,
    key_id: str,
    old_tags: List[Dict[str, Any]],
    new_tags: List[Dict[str, Any]],
):
    """

    Args:
        hub:
        ctx:
        key_id: aws kms key id
        old_tags: list of old tags
        new_tags: list of new tags

    Returns:
        {"result": True|False, "comment": Tuple, "ret": "Tags after update"}

    """
    result = dict(comment=(), result=True, ret=None)

    tags_to_add = list()
    old_tags_map = {tag.get("TagKey"): tag for tag in old_tags}
    tags_result = copy.deepcopy(old_tags_map)
    for tag in new_tags:
        if tag.get("TagKey") in old_tags_map:
            # Make sure the key and value are the same before deleting
            if tag.get("TagValue") == old_tags_map.get(tag.get("TagKey")).get(
                "KeyValue"
            ):
                del old_tags_map[tag.get("TagKey")]
            else:
                tags_to_add.append(tag)
        else:
            tags_to_add.append(tag)
    tags_to_remove = [tag.get("TagKey") for tag in old_tags_map.values()]
    if tags_to_remove and not ctx.get("test", False):
        delete_ret = await hub.exec.boto3.client.kms.untag_resource(
            ctx, KeyId=key_id, TagKeys=tags_to_remove
        )
        if not delete_ret["result"]:
            result["comment"] = delete_ret["comment"]
            result["result"] = False
            return result
    if tags_to_add and not ctx.get("test", False):
        add_ret = await hub.exec.boto3.client.kms.tag_resource(
            ctx, KeyId=key_id, Tags=tags_to_add
        )
        if not add_ret["result"]:
            result["comment"] = add_ret["comment"]
            result["result"] = False
            return result
    for key in tags_to_remove:
        tags_result.pop(key)
    result["ret"] = {"tags": list(tags_result.values()) + tags_to_add}
    if ctx.get("test", False):
        result["comment"] = (
            f"Would update tags: Add [{tags_to_add}] Remove [{tags_to_remove}]",
        )
    else:
        result["comment"] = (
            f"Updated tags: Added [{tags_to_add}] Removed [{tags_to_remove}]",
        )
    return result



import copy
from typing import Any
from typing import Dict
from typing import List


async def update_tags(
    hub,
    ctx,
    resource_id,
    old_tags: List[Dict[str, Any]],
    new_tags: List[Dict[str, Any]],
):
    """
    Update tags of AWS EC2 resources
    Args:
        hub:
        ctx:
        resource_id: aws resource id
        old_tags: list of old tags
        new_tags: list of new tags

    Returns:
        {"result": True|False, "comment": A message Tuple, "ret": None}

    """
    tags_to_add = []
    tags_to_remove = []
    old_tags_map = {tag.get("Key"): tag for tag in old_tags or []}
    tags_result = copy.deepcopy(old_tags_map)
    if new_tags is not None:
        for tag in new_tags:
            if tag.get("Key") in old_tags_map:
                if tag.get("Value") == old_tags_map.get(tag.get("Key")).get("Value"):
                    del old_tags_map[tag.get("Key")]
                else:
                    tags_to_add.append(tag)
            else:
                tags_to_add.append(tag)
        tags_to_remove = [{"Key": tag.get("Key")} for tag in old_tags_map.values()]
    result = dict(comment=(), result=True, ret=None)
    if (not tags_to_remove) and (not tags_to_add):
        return result
    if tags_to_remove:
        if not ctx.get("test", False):
            delete_ret = await hub.exec.boto3.client.ec2.delete_tags(
                ctx, Resources=[resource_id], Tags=tags_to_remove
            )
            if not delete_ret["result"]:
                result["comment"] = delete_ret["comment"]
                result["result"] = False
                return result
        [tags_result.pop(key.get("Key"), None) for key in tags_to_remove]
    if tags_to_add:
        if not ctx.get("test", False):
            add_ret = await hub.exec.boto3.client.ec2.create_tags(
                ctx, Resources=[resource_id], Tags=tags_to_add
            )
            if not add_ret["result"]:
                result["comment"] = add_ret["comment"]
                result["result"] = False
                return result
    result["ret"] = {"tags": list(tags_result.values()) + tags_to_add}
    result["comment"] = (f"Update tags: Add [{tags_to_add}] Remove [{tags_to_remove}]",)
    return result



from typing import List


async def update_associations_routes(
    hub,
    ctx,
    route_table_id: str,
    old_routes: List = None,
    new_routes: List = None,
    old_associations: List = None,
    new_associations: List = None,
    old_propagating_vgws: List = None,
    new_propagating_vgws: List = None,
):
    """
    Update associations, routes and route propagations of a route table. This function compares the existing(old)
    associations or routes or route propagations of route table and the new associations or routes or route propagations.
    associations or routes or route propagations that are in the new route table but not in the old
    route table will be associated to route table. associations or routes or route propagations that are in the
    old route table  but not in the new route table will be disassociated from transit route table.

    Args:
        hub:
        ctx:
        route_table_id(Text): The AWS resource id of the existing route table
        old_routes(List):  routes of existing route table
        new_routes(List): new routes to update
        old_associations(List): associations of existing route table
        new_associations(List): new associations to be updated
        old_propagating_vgws(List): propagating virtual gateways of existing route table
        new_propagating_vgws(List): new propagating virtual gateways to be updated

    Returns:
        {"result": True|False, "comment": "A message", "ret": None}

    """
    result = dict(comment=(), result=True, ret=None)

    # compare old_routes and new_routes and update if routes are modified
    if new_routes is []:
        result["result"] = False
        result["comment"] = (
            "Route Table routes cannot be None. There should be at least one route in Route Table",
        )
        return result
    elif new_routes is not None:
        routes_to_modify = (
            hub.tool.aws.ec2.route_table_utils.get_route_table_routes_modifications(
                old_routes, new_routes
            )
        )
        if not routes_to_modify["result"]:
            result["comment"] = routes_to_modify["comment"]
            result["result"] = False
            return result
        routes_to_add = routes_to_modify["routes_to_add"]
        routes_to_delete = routes_to_modify["routes_to_delete"]
        routes_to_replace = routes_to_modify["routes_to_replace"]

        if not ctx.get("test", False):
            if routes_to_delete:
                for route_to_delete in routes_to_delete:
                    ret = await hub.exec.boto3.client.ec2.delete_route(
                        ctx, RouteTableId=route_table_id, **route_to_delete
                    )
                    if not ret.get("result"):
                        result["comment"] = result["comment"] + ret["comment"]
                        result["result"] = False
                        return result
                result["comment"] = result["comment"] + (
                    f"Deleted Routes: {routes_to_delete}, ",
                )
            if routes_to_replace:
                for route_to_replace in routes_to_replace:
                    ret = await hub.exec.boto3.client.ec2.replace_route(
                        ctx, RouteTableId=route_table_id, **route_to_replace
                    )
                    if not ret.get("result"):
                        result["comment"] = result["comment"] + ret["comment"]
                        result["result"] = False
                        return result
                result["comment"] = result["comment"] + (
                    f"Replace Routes: {routes_to_replace}, ",
                )
            if routes_to_add:
                for route_to_add in routes_to_add:
                    ret = await hub.exec.boto3.client.ec2.create_route(
                        ctx, RouteTableId=route_table_id, **route_to_add
                    )
                    if not ret.get("result"):
                        result["comment"] = result["comment"] + ret["comment"]
                        result["result"] = False
                        return result
                result["comment"] = result["comment"] + (
                    f"Added Routes: {routes_to_add}, ",
                )
    # If associations are None, we'll skip updating associations.
    if new_associations is not None:
        # compare old associations and new associations and update if associations are modified
        associations_to_modify = hub.tool.aws.ec2.route_table_utils.get_route_table_associations_modifications(
            old_associations, new_associations
        )
        if not associations_to_modify["result"]:
            result["comment"] = result["comment"] + associations_to_modify["comment"]
            return result
        associations_to_add = associations_to_modify["associations_to_add"]
        associations_to_delete = associations_to_modify["associations_to_delete"]

        if not ctx.get("test", False):
            if associations_to_delete:
                for association_to_delete in associations_to_delete:
                    ret = await hub.exec.boto3.client.ec2.disassociate_route_table(
                        ctx, **association_to_delete
                    )
                    if not ret.get("result"):
                        result["comment"] = result["comment"] + ret["comment"]
                        result["result"] = False
                        return result
                result["comment"] = result["comment"] + (
                    f"Deleted Associations: {associations_to_delete}, ",
                )
            if associations_to_add:
                for association_to_add in associations_to_add:
                    ret = await hub.exec.boto3.client.ec2.associate_route_table(
                        ctx, RouteTableId=route_table_id, **association_to_add
                    )
                    if not ret.get("result"):
                        result["comment"] = result["comment"] + ret["comment"]
                        result["result"] = False
                        return result
                result["comment"] = result["comment"] + (
                    f"Added Associations: {associations_to_add}, ",
                )
    if new_propagating_vgws is not None:
        # compare old and new propagating virtual gateways, update if propagating virtual gateways are modified
        propagating_vgws_to_modify = hub.tool.aws.ec2.route_table_utils.get_route_table_propagating_vgws_modifications(
            old_propagating_vgws, new_propagating_vgws
        )
        propagating_vgws_to_add = propagating_vgws_to_modify["propagating_vgws_to_add"]
        propagating_vgws_to_delete = propagating_vgws_to_modify[
            "propagating_vgws_to_delete"
        ]
        if not ctx.get("test", False):
            if propagating_vgws_to_delete:
                for propagating_vgw_to_delete in propagating_vgws_to_delete:
                    ret = await hub.exec.boto3.client.ec2.disable_vgw_route_propagation(
                        ctx,
                        RouteTableId=route_table_id,
                        GatewayId=propagating_vgw_to_delete,
                    )
                    if not ret.get("result"):
                        result["comment"] = result["comment"] + ret["comment"]
                        result["result"] = False
                        return result
                result["comment"] = result["comment"] + (
                    f"Deleted Propagating Virtual Gateways: {propagating_vgws_to_delete}, ",
                )
            if propagating_vgws_to_add:
                for propagating_vgw_to_add in propagating_vgws_to_add:
                    ret = await hub.exec.boto3.client.ec2.enable_vgw_route_propagation(
                        ctx,
                        RouteTableId=route_table_id,
                        GatewayId=propagating_vgw_to_add,
                    )
                    if not ret.get("result"):
                        result["comment"] = result["comment"] + ret["comment"]
                        result["result"] = False
                        return result
                result["comment"] = result["comment"] + (
                    f"Added Propagating Virtual Gateways: {propagating_vgws_to_add}, ",
                )

    if result["comment"]:
        result["comment"] = result["comment"] + (
            f"Updated route table {route_table_id}",
        )
    result["ret"] = {
        "routes": new_routes,
        "associations": new_associations,
        "vgws": new_propagating_vgws,
    }
    return result



import copy
from typing import Any
from typing import Dict
from typing import List


async def update_tags(
    hub,
    ctx,
    resource_arn: str,
    old_tags: List[Dict[str, Any]],
    new_tags: List[Dict[str, Any]],
):
    """

    Args:
        hub:
        ctx:
        resource_arn: aws elasticache arn
        old_tags: List of existing tags
        new_tags: List of new tags

    Returns:
        {"result": True|False, "comment": "A message", "ret": None}

    """

    result = dict(comment="", result=True, ret=None)

    old_tags_map = {tag.get("Key"): tag.get("Value") for tag in old_tags or []}
    new_tags_map = {tag.get("Key"): tag.get("Value") for tag in new_tags or []}
    tags_result = copy.deepcopy(old_tags_map)

    if old_tags_map == new_tags_map:
        result["comment"] = "All tags are updated!"
        return result

    tags_to_add = []
    tags_to_delete = []

    for key, value in new_tags_map.items():
        if (key in old_tags_map and old_tags_map.get(key) != new_tags_map.get(key)) or (
            key not in old_tags_map
        ):
            tags_to_add.append({"Key": key, "Value": value})

    for key in old_tags_map:
        if key not in new_tags_map:
            tags_to_delete.append(key)
    try:
        if not ctx.get("test", False) and tags_to_delete:
            delete_tag_resp = (
                await hub.exec.boto3.client.elasticache.remove_tags_from_resource(
                    ctx, ResourceName=resource_arn, TagKeys=tags_to_delete
                )
            )
            if not delete_tag_resp["result"]:
                hub.log.debug(f"Could not delete tags {tags_to_delete}")
                result["comment"] = delete_tag_resp["comment"]
                result["result"] = False
                return result

            hub.log.debug(f"Deleted tags {tags_to_delete}")
        [tags_result.pop(key, None) for key in tags_to_delete]
    except hub.tool.boto3.exception.ClientError as e:
        hub.log.debug(f"Error while deleting tags {tags_to_delete}")
        result["comment"] = f"{e.__class__.__name__}: {e}"
        result["result"] = False

    try:
        if not ctx.get("test", False) and tags_to_add:
            create_tag_resp = (
                await hub.exec.boto3.client.elasticache.add_tags_to_resource(
                    ctx, ResourceName=resource_arn, Tags=tags_to_add
                )
            )
            if not create_tag_resp["result"]:
                hub.log.debug(f"Could not create tags {tags_to_add}")
                result["comment"] = create_tag_resp["comment"]
                result["result"] = False
                return result

            hub.log.debug(f"Created tags {tags_to_add}")
    except hub.tool.boto3.exception.ClientError as e:
        hub.log.debug(f"Error while creating tags {tags_to_add}")
        result["comment"] = f"{e.__class__.__name__}: {e}"
        result["result"] = False

    result["ret"] = {"tags": list(tags_result.values()) + tags_to_add}
    result["comment"] = (f"Update tags: Add [{tags_to_add}] Remove [{tags_to_delete}]",)
    result["result"] = True
    return result



from collections import OrderedDict
from typing import Any
from typing import Dict
from typing import List


def convert_raw_hosted_zone_to_present(
    hub,
    raw_resource: Dict[str, Any],
    idem_resource_name: str = None,
    tags: List = None,
) -> Dict[str, Any]:

    """
    Given an object state from aws, this function creates a translated resource object in response.

    Args:
        hub: required for functions in hub
        raw_resource (Dict): The dictionary object from where the raw state of resource needs to be translated.
        idem_resource_name (Text): The idem of the idem resource
        tags (List): The tags.
                * Key (Text) -- The key of the tag. Tag keys are case-sensitive and accept a maximum of 127 Unicode characters. May not begin with aws: .
                * Value (Text) -- The value of the tag. Tag values are case-sensitive and accept a maximum of 255 Unicode characters.


    Returns: Dict[str, Any]
    """

    hosted_zone = raw_resource["ret"].get("HostedZone")
    vpcs = raw_resource["ret"].get("VPCs")
    delegation_set = raw_resource["ret"].get("DelegationSet")
    resource_id = hosted_zone.get("Id")
    resource_parameters = OrderedDict(
        {
            "Name": "hosted_zone_name",
            "CallerReference": "caller_reference",
            "Id": "resource_id",
        }
    )
    resource_translated = {"name": idem_resource_name, "resource_id": resource_id}
    for parameter_raw, parameter_present in resource_parameters.items():
        if parameter_raw in hosted_zone:
            resource_translated[parameter_present] = hosted_zone.get(parameter_raw)
    resource_translated["config"] = hosted_zone.get("Config").copy()
    if vpcs is not None:
        resource_translated["vpcs"] = vpcs
    if delegation_set is not None:
        resource_translated["delegation_set"] = delegation_set.copy()
    resource_translated["tags"] = tags

    return resource_translated



from typing import Any
from typing import Dict


async def update_eks_tags(
    hub,
    ctx,
    resource_arn,
    old_tags: Dict[str, Any],
    new_tags: Dict[str, Any],
):
    """

    Args:
        hub:
        ctx:
        resource_arn: aws resource arn
        old_tags: Dict of old tags
        new_tags: Dict of new tags

    Returns:
        {"result": True|False, "comment": "A message", "ret": None}

    """
    tags_to_add = {}
    tags_to_remove = []
    if new_tags is not None:
        for tag in new_tags:
            if new_tags[tag] == old_tags.get(tag):
                del old_tags[tag]
            else:
                tags_to_add.update({tag: new_tags.get(tag)})
        tags_to_remove = list(old_tags.keys())

    result = dict(comment="", result=True, ret=None)
    if tags_to_remove:
        delete_ret = await hub.exec.boto3.client.eks.untag_resource(
            ctx, resourceArn=resource_arn, tagKeys=tags_to_remove
        )
        if not delete_ret["result"]:
            result["comment"] = delete_ret["comment"]
            result["result"] = False
            return result
        result["ret"] = delete_ret["ret"]
    if tags_to_add:
        add_ret = await hub.exec.boto3.client.eks.tag_resource(
            ctx, resourceArn=resource_arn, tags=tags_to_add
        )
        if not add_ret["result"]:
            result["comment"] = add_ret["comment"]
            result["result"] = False
            return result
        result["ret"] = add_ret["ret"]

    result["comment"] = f"Update tags: Add [{tags_to_add}] Remove [{tags_to_remove}]"
    return result



"""
Autogenerated state module using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

Note..

The describe function uses the describe_spot_instance_requests to request spot instance details.
The present function uses the request_spot_instances to request creation of a spot instance.
The delete function uses the standard terminate_instances as the cancel_spot_requests does
not terminate an instance that comes online.

If you wish to delete or terminate an instance, and it's a spot instance, you must use the describe
function here to obtain the InstanceID and then use ec2.instance.absent to fully terminate the instance.
"""
from collections import OrderedDict
from typing import Any
from typing import Dict
from typing import List


__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    spot_instance_request_id=None,
    availability_zone_group=None,
    instance_count=None,
    instance_interruption_behavior=None,
    launch_group=None,
    launch_specification=None,
    spot_price=None,
    type=None,
    valid_from=None,
    valid_until=None,
    tags: List = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Request Parameters
        The following parameters are for this specific action. For more information about required
         and optional parameters that are common to all actions, see Common Query Parameters.

        spot_instance_id(Text):
            The spot instance request id

        availability_zone_group
            The user-specified name for a logical grouping of requests.

            When you specify an Availability Zone group in a Spot Instance request,
            all Spot Instances in the request are launched in the same Availability Zone.
            Instance proximity is maintained with this parameter, but the choice of
            Availability Zone is not. The group applies only to requests for Spot Instances
            of the same instance type. Any additional Spot Instance requests that are
            specified with the same Availability Zone group name are launched in that
            same Availability Zone, as long as at least one instance from the group is still active.

            If there is no active instance running in the Availability Zone group that
            you specify for a new Spot Instance request (all instances are terminated,
            the request is expired, or the maximum price you specified falls below
            current Spot price), then Amazon EC2 launches the instance in any Availability
            Zone where the constraint can be met. Consequently, the subsequent set of Spot
            Instances could be placed in a different zone from the original request,
            even if you specified the same Availability Zone group.

            Default: Instances are launched in any available Availability Zone.

            Type: String

            Required: No

        BlockDurationMinutes
            Deprecated.

            Type: Integer

            Required: No

        instance_count
            The maximum number of Spot Instances to launch.

            Default: 1

            Type: Integer

            Required: No

        instance_interruption_behavior
            The behavior when a Spot Instance is interrupted. The default is terminate.

            Type: String

            Valid Values: hibernate | stop | terminate

            Required: No

        launch_group
            The instance launch group. Launch groups are Spot Instances that launch together and terminate together.

            Default: Instances are launched and terminated individually

            Type: String

            Required: No

        launch_specification
            The launch specification.

            Type: RequestSpotLaunchSpecification object dict

            Required: No

            Below are the attributes available for LaunchSpecification:

                SecurityGroupIds (list) --
                One or more security group IDs.

                (string) --
                SecurityGroups (list) --
                One or more security groups. When requesting instances in a VPC,
                you must specify the IDs of the security groups. When requesting
                instances in EC2-Classic, you can specify the names or the IDs
                of the security groups.

                (string) --
                AddressingType (string) --
                Deprecated.

                BlockDeviceMappings (list) --
                One or more block device mapping entries. You can't specify both
                a snapshot ID and an encryption value. This is because only blank
                volumes can be encrypted on creation. If a snapshot is the basis
                for a volume, it is not blank and its encryption status is used
                for the volume encryption status.

                (dict) --
                Describes a block device mapping, which defines the EBS volumes
                and instance store volumes to attach to an instance at launch.

                DeviceName (string) --
                The device name (for example, /dev/sdh or xvdh ).

                VirtualName (string) --
                The virtual device name (ephemeral N). Instance store volumes
                are numbered starting from 0. An instance type with 2 available
                instance store volumes can specify mappings for ephemeral0 and
                ephemeral1. The number of available instance store volumes depends
                on the instance type. After you connect to the instance,
                you must mount the volume.

                NVMe instance store volumes are automatically enumerated and
                assigned a device name. Including them in your block device
                mapping has no effect.

                Constraints: For M3 instances, you must specify instance store
                volumes in the block device mapping for the instance. When you
                launch an M3 instance, we ignore any instance store volumes
                specified in the block device mapping for the AMI.

                Ebs (dict) --
                Parameters used to automatically set up EBS volumes when the
                instance is launched.

                DeleteOnTermination (boolean) --
                Indicates whether the EBS volume is deleted on instance termination. For more information, see Preserving Amazon EBS volumes on instance termination in the Amazon EC2 User Guide .

                Iops (integer) --
                The number of I/O operations per second (IOPS). For gp3 , io1 , and io2 volumes, this represents the number of IOPS that are provisioned for the volume. For gp2 volumes, this represents the baseline performance of the volume and the rate at which the volume accumulates I/O credits for bursting.

                The following are the supported values for each volume type:

                gp3 : 3,000-16,000 IOPS
                io1 : 100-64,000 IOPS
                io2 : 100-64,000 IOPS
                For io1 and io2 volumes, we guarantee 64,000 IOPS only for Instances built on the Nitro System . Other instance families guarantee performance up to 32,000 IOPS.

                This parameter is required for io1 and io2 volumes. The default for gp3 volumes is 3,000 IOPS. This parameter is not supported for gp2 , st1 , sc1 , or standard volumes.

                SnapshotId (string) --
                The ID of the snapshot.

                VolumeSize (integer) --
                The size of the volume, in GiBs. You must specify either a snapshot ID or a volume size. If you specify a snapshot, the default is the snapshot size. You can specify a volume size that is equal to or larger than the snapshot size.

                The following are the supported volumes sizes for each volume type:

                gp2 and gp3 :1-16,384
                io1 and io2 : 4-16,384
                st1 and sc1 : 125-16,384
                standard : 1-1,024
                VolumeType (string) --
                The volume type. For more information, see Amazon EBS volume types in the Amazon EC2 User Guide . If the volume type is io1 or io2 , you must specify the IOPS that the volume supports.

                KmsKeyId (string) --
                Identifier (key ID, key alias, ID ARN, or alias ARN) for a customer managed CMK under which the EBS volume is encrypted.

                This parameter is only supported on BlockDeviceMapping objects called by RunInstances , RequestSpotFleet , and RequestSpotInstances .

                Throughput (integer) --
                The throughput that the volume supports, in MiB/s.

                This parameter is valid only for gp3 volumes.

                Valid Range: Minimum value of 125. Maximum value of 1000.

                OutpostArn (string) --
                The ARN of the Outpost on which the snapshot is stored.

                Encrypted (boolean) --
                Indicates whether the encryption state of an EBS volume is changed while being restored from a backing snapshot. The effect of setting the encryption state to true depends on the volume origin (new or from a snapshot), starting encryption state, ownership, and whether encryption by default is enabled. For more information, see Amazon EBS encryption in the Amazon EC2 User Guide .

                In no case can you remove encryption from an encrypted volume.

                Encrypted volumes can only be attached to instances that support Amazon EBS encryption. For more information, see Supported instance types .

                This parameter is not returned by .

                NoDevice (string) --
                To omit the device from the block device mapping, specify an empty string. When this property is specified, the device is removed from the block device mapping regardless of the assigned value.

                EbsOptimized (boolean) --
                Indicates whether the instance is optimized for EBS I/O. This optimization provides dedicated throughput to Amazon EBS and an optimized configuration stack to provide optimal EBS I/O performance. This optimization isn't available with all instance types. Additional usage charges apply when using an EBS Optimized instance.

                Default: false

                IamInstanceProfile (dict) --
                The IAM instance profile.

                Arn (string) --
                The Amazon Resource Name (ARN) of the instance profile.

                Name (string) --
                The name of the instance profile.

                ImageId (string) --
                The ID of the AMI.

                InstanceType (string) --
                The instance type.

                KernelId (string) --
                The ID of the kernel.

                KeyName (string) --
                The name of the key pair.

                Monitoring (dict) --
                Indicates whether basic or detailed monitoring is enabled for the instance.

                Default: Disabled

                Enabled (boolean) -- [REQUIRED]
                Indicates whether detailed monitoring is enabled. Otherwise, basic monitoring is enabled.

                NetworkInterfaces (list) --
                One or more network interfaces. If you specify a network interface, you must specify subnet IDs and security group IDs using the network interface.

                (dict) --
                Describes a network interface.

                AssociatePublicIpAddress (boolean) --
                Indicates whether to assign a public IPv4 address to an instance you launch in a VPC. The public IP address can only be assigned to a network interface for eth0, and can only be assigned to a new network interface, not an existing one. You cannot specify more than one network interface in the request. If launching into a default subnet, the default value is true .

                DeleteOnTermination (boolean) --
                If set to true , the interface is deleted when the instance is terminated. You can specify true only if creating a new network interface when launching an instance.

                Description (string) --
                The description of the network interface. Applies only if creating a network interface when launching an instance.

                DeviceIndex (integer) --
                The position of the network interface in the attachment order. A primary network interface has a device index of 0.

                If you specify a network interface when launching an instance, you must specify the device index.

                Groups (list) --
                The IDs of the security groups for the network interface. Applies only if creating a network interface when launching an instance.

                (string) --
                Ipv6AddressCount (integer) --
                A number of IPv6 addresses to assign to the network interface. Amazon EC2 chooses the IPv6 addresses from the range of the subnet. You cannot specify this option and the option to assign specific IPv6 addresses in the same request. You can specify this option if you've specified a minimum number of instances to launch.

                Ipv6Addresses (list) --
                One or more IPv6 addresses to assign to the network interface. You cannot specify this option and the option to assign a number of IPv6 addresses in the same request. You cannot specify this option if you've specified a minimum number of instances to launch.

                (dict) --
                Describes an IPv6 address.

                Ipv6Address (string) --
                The IPv6 address.

                NetworkInterfaceId (string) --
                The ID of the network interface.

                If you are creating a Spot Fleet, omit this parameter because you can’t specify a network interface ID in a launch specification.

                PrivateIpAddress (string) --
                The private IPv4 address of the network interface. Applies only if creating a network interface when launching an instance. You cannot specify this option if you're launching more than one instance in a RunInstances request.

                PrivateIpAddresses (list) --
                One or more private IPv4 addresses to assign to the network interface. Only one private IPv4 address can be designated as primary. You cannot specify this option if you're launching more than one instance in a RunInstances request.

                (dict) --
                Describes a secondary private IPv4 address for a network interface.

                Primary (boolean) --
                Indicates whether the private IPv4 address is the primary private IPv4 address. Only one IPv4 address can be designated as primary.

                PrivateIpAddress (string) --
                The private IPv4 addresses.

                SecondaryPrivateIpAddressCount (integer) --
                The number of secondary private IPv4 addresses. You can't specify this option and specify more than one private IP address using the private IP addresses option. You cannot specify this option if you're launching more than one instance in a RunInstances request.

                SubnetId (string) --
                The ID of the subnet associated with the network interface. Applies only if creating a network interface when launching an instance.

                AssociateCarrierIpAddress (boolean) --
                Indicates whether to assign a carrier IP address to the network interface.

                You can only assign a carrier IP address to a network interface that is in a subnet in a Wavelength Zone. For more information about carrier IP addresses, see Carrier IP addresses in the Amazon Web Services Wavelength Developer Guide.

                InterfaceType (string) --
                The type of network interface.

                To create an Elastic Fabric Adapter (EFA), specify efa . For more information, see Elastic Fabric Adapter in the Amazon Elastic Compute Cloud User Guide .

                Valid values: interface | efa

                NetworkCardIndex (integer) --
                The index of the network card. Some instance types support multiple network cards. The primary network interface must be assigned to network card index 0. The default is network card index 0.

                If you are using RequestSpotInstances to create Spot Instances, omit this parameter because you can’t specify the network card index when using this API. To specify the network card index, use RunInstances .

                Ipv4Prefixes (list) --
                One or more IPv4 delegated prefixes to be assigned to the network interface. You cannot use this option if you use the Ipv4PrefixCount option.

                (dict) --
                Describes the IPv4 prefix option for a network interface.

                Ipv4Prefix (string) --
                The IPv4 prefix. For information, see Assigning prefixes to Amazon EC2 network interfaces in the Amazon Elastic Compute Cloud User Guide .

                Ipv4PrefixCount (integer) --
                The number of IPv4 delegated prefixes to be automatically assigned to the network interface. You cannot use this option if you use the Ipv4Prefix option.

                Ipv6Prefixes (list) --
                One or more IPv6 delegated prefixes to be assigned to the network interface. You cannot use this option if you use the Ipv6PrefixCount option.

                (dict) --
                Describes the IPv4 prefix option for a network interface.

                Ipv6Prefix (string) --
                The IPv6 prefix.

                Ipv6PrefixCount (integer) --
                The number of IPv6 delegated prefixes to be automatically assigned to the network interface. You cannot use this option if you use the Ipv6Prefix option.

                Placement (dict) --
                The placement information for the instance.

                AvailabilityZone (string) --
                The Availability Zone.

                [Spot Fleet only] To specify multiple Availability Zones, separate them using commas; for example, "us-west-2a, us-west-2b".

                GroupName (string) --
                The name of the placement group.

                Tenancy (string) --
                The tenancy of the instance (if the instance is running in a VPC). An instance with a tenancy of dedicated runs on single-tenant hardware. The host tenancy is not supported for Spot Instances.

                RamdiskId (string) --
                The ID of the RAM disk.

                SubnetId (string) --
                The ID of the subnet in which to launch the instance.

                UserData (string) --
                The Base64-encoded user data for the instance. User data is limited to 16 KB.

        spot_price
            The maximum price per hour that you are willing to pay for
            a Spot Instance. The default is the On-Demand price.

            Type: String

            Required: No

        tag_specifications
            The key-value pair for tagging the Spot Instance request on creation.
            The value for ResourceType must be spot-instances-request, otherwise
            the Spot Instance request fails. To tag the Spot Instance request
            after it has been created, see CreateTags.

            Type: Array of TagSpecification objects

            tags(List, optional): The tags to assign to the VPC. Defaults to None.
                * Key (string) -- The key of the tag.
                Tag keys are case-sensitive and accept a maximum of 127 Unicode characters.
                May not begin with aws.
                * Value (string) -- The value of the tag. Tag values are case-sensitive
                and accept a maximum of 255 Unicode characters.

            Required: No

        type
            The Spot Instance request type.

            Default: one-time

            Type: String

            Valid Values: one-time | persistent

            Required: No

        valid_from
            The start date of the request. If this is a one-time request,
            the request becomes active at this date and time and remains
            active until all instances launch, the request expires, or
            the request is canceled. If the request is persistent, the
            request becomes active at this date and time and remains
            active until it expires or is canceled.

            The specified start date and time cannot be equal to the current
            date and time. You must specify a start date and time that
            occurs after the current date and time.

            Type: Timestamp

            Required: No

        valid_until
            The end date of the request, in UTC format (YYYY-MM-DDTHH:MM:SSZ).

            For a persistent request, the request remains active until the ValidUntil
            date and time is reached. Otherwise, the request remains active until you cancel it.

            For a one-time request, the request remains active until all instances launch,
            the request is canceled, or the ValidUntil date and time is reached.
            By default, the request is valid for 7 days from the date the request
            was created.

            Type: Timestamp

            Required: No

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            resource_is_present:
              aws.ec2.spot_instance.present:
                - name: value

        .. code-block:: sls

            resource_is_present:
              aws.ec2.spot_instance.present:
                instance_count: 5
                launch_specification:
                    ImageId: ami-fce3c696
                    KeyName: awskey.pem
                    SecurityGroups: ['sg-709f8709']
                    InstanceType: m4.large
                    Placement:
                        AvailabilityZone: us-east-1a
                    BlockDeviceMappings:
                            Ebs:
                                SnapshotId: snap-f70deff0
                                VolumeSize: 100
                                DeleteOnTermination: True
                                VolumeType: gp2
                                Iops: 300
                                Encrypted: False
                    EbsOptimized: True
                    Monitoring:
                        Enabled: True
                    SecurityGroupIds: sg-709f8709
                spot_price: 0.03
                type: one-time
    """
    if spot_instance_request_id is None:
        spot_instance_request_id = name

    result = dict(comment="", old_state=None, new_state=None, name=name, result=True)
    before = await hub.exec.boto3.client.ec2.describe_spot_instance_requests(
        spot_instance_request_id
    )
    result["new_state"] = before

    if ctx.get("test", False):
        if before:
            result["comment"] = "Would update aws.ec2.spot_instance"
            result["result"] = True
        else:
            result["comment"] = "Would create aws.ec2.spot_instance"
            result["result"] = True
        return result

    if not before:
        try:
            # Although it's not specified in the API documentation,
            # apparently 'SecurityGroups' parameter requires the
            # names of the security groups, not the IDs, inside the
            # LaunchSpecification dict
            ret = await hub.exec.boto3.client.ec2.request_spot_instances(
                ctx,
                DryRun=False,
                **{
                    "AvailabilityZoneGroup": availability_zone_group,
                    "InstanceCount": instance_count,
                    "LaunchGroup": launch_group,
                    "LaunchSpecification": launch_specification,
                    "SpotPrice": spot_price,
                    "Type": type,
                    "ValidFrom": valid_from,
                    "ValidUntil": valid_until,
                    "TagSpecifications": [{"ResourceType": "instance", "Tags": tags}]
                    if tags
                    else None,
                    "InstanceInterruptionBehavior": instance_interruption_behavior,
                },
            )
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = f"{e.__class__.__name__}: {e}"
            result["result"] = False
        else:
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                return result
            assert ret["result"] is True
            assert (
                ret["ret"]["SpotInstanceRequests"][0]["Status"]["Code"]
                == "pending-evaluation"
            )
            name = ret["ret"]["SpotInstanceRequests"][0]["SpotInstanceRequestId"]
            # Since we don't have a valid instance yet use the RequestId as the name for lookup
            result["comment"] = name
    else:
        result["comment"] = f"{spot_instance_request_id} already exists."

    if not before:
        try:
            after = await hub.exec.boto3.client.ec2.describe_spot_instance_requests(
                spot_instance_request_id
            )
        except Exception as e:
            result["comment"] = str(e)
            result["result"] = False
        else:
            result["new_state"] = after
            result["old_state"] = before

    return result


async def absent(
    hub, ctx, name: str, spot_instance_request_id: str = None
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Cancels one or more Spot Instance requests.

    Args:
        spot_instance_request_id(Text): The spot instance request id

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            resource_is_absent:
              aws.ec2.spot_instance.absent:
                - name: value
    """
    if spot_instance_request_id is None:
        spot_instance_request_id = name
    result = dict(comment="", old_state=None, new_state=None, name=name, result=True)
    before = await hub.exec.boto3.client.ec2.describe_spot_instance_requests(
        spot_instance_request_id
    )

    if not before:
        result["comment"] = f"'{name}' already absent"
    elif ctx.get("test", False):
        result["comment"] = f"Would cancel aws.ec2.request_spot_instances {name}"
        return result
    else:
        try:
            # Warning...
            # Canceling a Spot Instance request does not terminate
            # running Spot Instances associated with the request
            ret = await hub.exec.boto3.client.ec2.cancel_spot_instance_requests(
                ctx,
                DryRun=False,
                **{"SpotInstanceRequestIds": [spot_instance_request_id]},
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                return result
            result["comment"] = f"Deleted '{name}'"
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = f"{e.__class__.__name__}: {e}"
        result["old_state"] = before

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    result = {}
    ret = await hub.exec.boto3.client.ec2.describe_spot_instance_requests(
        ctx, MaxResults=5
    )
    if not ret["result"]:
        hub.log.debug(f"Could not describe spot instances {ret['comment']}")
        return {}

    instances = []
    for reservation in ret["ret"]["SpotInstanceRequests"]:
        instances.append(reservation)
    # Primary parameters for spot instances
    describe_parameters = OrderedDict(
        {
            "CreateTime": "create_time",
            "LaunchedAvailabilityZone": "launched_availability_zone",
            "ProductDescription": "product_description",
            "SpotInstanceRequestId": "spot_instance_request_id",
            "SpotPrice": "spot_price",
            "State": "state",
            "Status": "status",
            "Tags": "tags",
        }
    )

    for instance in instances:
        instance_id = instance.get("InstanceId")
        # Instance specifications described below
        describe_parameters_specs = OrderedDict(
            {
                "BlockDeviceMappings": "block_device_mappings",
                "ImageId": "image_id",
                "InstanceType": "instance_type",
                "Monitoring": "monitoring",
                "Placement": "placement",
                "NetworkInterfaces": "network_interfaces",
                "SecurityGroups": "security_groups",
            }
        )
        instance_translated = []
        for p_old_key, p_new_key in describe_parameters.items():
            if instance.get(p_old_key) is not None:
                instance_translated.append({p_new_key: instance.get(p_old_key)})

        for parameter_old_key, parameter_new_key in describe_parameters_specs.items():
            instance_specs = instance.get("LaunchSpecification")
            if instance_specs.get(parameter_old_key) is not None:
                instance_translated.append(
                    {parameter_new_key: instance_specs.get(parameter_old_key)}
                )

        result[instance_id] = {"aws.ec2.spot_instance.present": instance_translated}
    return result



from typing import Any
from typing import Dict


async def update_ipv6_cidr_blocks(
    hub,
    ctx,
    subnet_id: str,
    old_ipv6_cidr_block: Dict[str, Any],
    new_ipv6_cidr_block: Dict[str, Any],
):
    """
    Update associated ipv6 cidr block of a subnet. This function supports associating an ipv6 cidr block, or updating
     the existing(old) ipv6 cidr block to the new ipv6 cidr block. Disassociating an ipv6 cidr block is not supported,
     due to how an sls file works currently. If ipv6_cidr_block parameter is left as blank, Idem-aws will do no-op
     on the subnet's ipv6 cidr block. To disassociate an ipv6 cidr block, a user will have to delete the subnet and
     re-create it without the ipv6 cidr block.

    Args:
        hub:
        ctx:
        subnet_id: The AWS resource id of the existing subnet
        old_ipv6_cidr_block: The ipv6 cidr block on the existing vpc
        new_ipv6_cidr_block: The expected ipv6 cidr block on the existing vpc

    Returns:
        {"result": True|False, "comment": A message Tuple, "ret": Dict}

    """
    result = dict(comment=(), result=True, ret=None)
    if old_ipv6_cidr_block is None and new_ipv6_cidr_block:
        if ctx.get("test", False):
            result["ret"] = {
                "ipv6_cidr_block": new_ipv6_cidr_block.get("Ipv6CidrBlock")
            }
            return result
        else:
            ret = await hub.exec.boto3.client.ec2.associate_subnet_cidr_block(
                ctx,
                SubnetId=subnet_id,
                Ipv6CidrBlock=new_ipv6_cidr_block.get("Ipv6CidrBlock"),
            )
            result["result"] = ret["result"]
            if result["result"]:
                hub.log.info(
                    f"Add subnet {subnet_id} ipv6 cidr block {new_ipv6_cidr_block.get('Ipv6CidrBlock')}"
                )
                result["ret"] = {
                    "ipv6_cidr_block": new_ipv6_cidr_block.get("Ipv6CidrBlock")
                }
            else:
                result["comment"] = ret["comment"]
            return result
    elif old_ipv6_cidr_block and new_ipv6_cidr_block:
        if old_ipv6_cidr_block.get("Ipv6CidrBlock") != new_ipv6_cidr_block.get(
            "Ipv6CidrBlock"
        ):
            if ctx.get("test", False):
                result["ret"] = {
                    "ipv6_cidr_block": new_ipv6_cidr_block.get("Ipv6CidrBlock")
                }
                return result
            else:
                ret = await hub.exec.boto3.client.ec2.disassociate_subnet_cidr_block(
                    ctx, AssociationId=old_ipv6_cidr_block.get("AssociationId")
                )
                if not ret.get("result"):
                    result["comment"] = ret["comment"]
                    result["result"] = False
                    return result
                ret = await hub.exec.boto3.client.ec2.associate_subnet_cidr_block(
                    ctx,
                    SubnetId=subnet_id,
                    Ipv6CidrBlock=new_ipv6_cidr_block.get("Ipv6CidrBlock"),
                )
                result["result"] = ret["result"]
                if result["result"]:
                    hub.log.info(
                        f"Update subnet {subnet_id} ipv6 cidr block from {old_ipv6_cidr_block.get('Ipv6CidrBlock')}"
                        f" to {new_ipv6_cidr_block.get('Ipv6CidrBlock')}"
                    )
                    result["ret"] = {
                        "ipv6_cidr_block": new_ipv6_cidr_block.get("Ipv6CidrBlock")
                    }
                    return result
                else:
                    result["comment"] = ret["comment"]
    return result



import copy
from typing import Any
from typing import Dict
from typing import List


async def update_instance_profile_tags(
    hub,
    ctx,
    instance_profile_name: str,
    old_tags: List[Dict[str, Any]],
    new_tags: List[Dict[str, Any]],
):
    """
    Update tags of AWS IAM Instance Profile

    TODO - this method might fail with localstack but is successful with a real AWS account

    Args:
        hub: The redistributed pop central hub.
        ctx: A dict with the keys/values for the execution of the Idem run located in
        `hub.idem.RUNS[ctx['run_name']]`.
        instance_profile_name: AWS IAM instance profile name
        old_tags: list of old tags
        new_tags: list of new tags

    Returns:
        {"result": True|False, "comment": Tuple, "ret": "Tags after update"}
    """
    result = dict(comment=(), result=True, ret=None)

    tags_to_add = list()
    old_tags_map = {tag.get("Key"): tag for tag in old_tags}
    tags_result = copy.deepcopy(old_tags_map)
    for tag in new_tags:
        if tag.get("Key") in old_tags_map:
            if tag.get("Value") == old_tags_map.get(tag.get("Key")).get("Value"):
                del old_tags_map[tag.get("Key")]
            else:
                tags_to_add.append(tag)
        else:
            tags_to_add.append(tag)
    tags_to_remove = [{"Key": tag.get("Key")} for tag in old_tags_map.values()]
    if tags_to_remove:
        if not ctx.get("test", False):
            delete_ret = await hub.exec.boto3.client.iam.untag_instance_profile(
                ctx, InstanceProfileName=instance_profile_name, TagKeys=tags_to_remove
            )
            if not delete_ret["result"]:
                result["comment"] = delete_ret["comment"]
                result["result"] = False
                return result
        [tags_result.pop(key.get("Key"), None) for key in tags_to_remove]
    if tags_to_add:
        if not ctx.get("test", False):
            add_ret = await hub.exec.boto3.client.iam.tag_instance_profile(
                ctx, InstanceProfileName=instance_profile_name, Tags=tags_to_add
            )
            if not add_ret["result"]:
                result["comment"] = add_ret["comment"]
                result["result"] = False
                return result

    result["ret"] = {"tags": list(tags_result.values()) + tags_to_add}
    result["comment"] = (
        f"Update tags on instance-profile: Add [{tags_to_add}] Remove [{tags_to_remove}]",
    )
    return result



"""
Autogenerated using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

hub.exec.boto3.client.dynamodb.create_table
hub.exec.boto3.client.dynamodb.delete_table
hub.exec.boto3.client.dynamodb.describe_table
hub.exec.boto3.client.dynamodb.list_tables
hub.exec.boto3.client.dynamodb.update_table
resource = hub.tool.boto3.resource.create(ctx, "dynamodb", "Table", name)
hub.tool.boto3.resource.exec(resource, delete, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, delete_item, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, get_item, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, put_item, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, query, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, scan, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, update, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, update_item, *args, **kwargs)
"""
import copy
from typing import Any
from typing import Dict
from typing import List


async def present(
    hub,
    ctx,
    name: str,
    attribute_definitions: List,
    key_schema: List,
    resource_id: str = None,
    billing_mode: str = None,
    sse_specification: dict = None,
    stream_specification: dict = None,
    provisioned_throughput: dict = None,
    local_secondary_indexes: List = None,
    global_secondary_indexes: List = None,
    replica_updates: List = None,
    tags: List = None,
    table_class: str = None,
    timeout: Dict = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    The CreateTable operation adds a new table to your account. In an Amazon Web Services account, table names must
    be unique within each Region. That is, you can have two tables with same name if you create the tables in
    different Regions.  CreateTable is an asynchronous operation. Upon receiving a CreateTable request, DynamoDB
    immediately returns a response with a TableStatus of CREATING. After the table is created, DynamoDB sets the
    TableStatus to ACTIVE. You can perform read and write operations only on an ACTIVE table.  You can optionally
    define secondary indexes on the new table, as part of the CreateTable operation. If you want to create multiple
    tables with secondary indexes on them, you must create the tables sequentially. Only one table with secondary
    indexes can be in the CREATING state at any given time. You can use the DescribeTable action to check the table
    status.

    Args:
        name(Text): A name of the table to create.
        resource_id(Text, optional): A name of the DynamoDB table
        key_schema(list): Specifies the attributes that make up the primary key for a table or an index. The attributes
            in KeySchema must also be defined in the AttributeDefinitions array.

            Each KeySchemaElement in the array is composed of:
                AttributeName - The name of this key attribute.
                KeyType - The role that the key attribute will assume:
                    HASH - partition key
                    RANGE - sort key
        attribute_definitions(List): An array of attributes that describe the key schema for the table and indexes.
        replica_updates: A list of replica update actions (create, delete, or update) for the table.
        table_class(Text, optional): The table class of the new table. Valid values are STANDARD and
                STANDARD_INFREQUENT_ACCESS.
        billing_mode(Text, optional): Controls how users are charged for read and write throughput and how they manage
            capacity.
                PROVISIONED - AWS recommends using PROVISIONED for predictable workloads. PROVISIONED sets the billing
                            mode to Provisioned Mode.
                PAY_PER_REQUEST - AWS recommends using PAY_PER_REQUEST for unpredictable workloads. PAY_PER_REQUEST
                            sets the billing mode to On-Demand Mode .
        global_secondary_indexes(list, optional): One or more global secondary indexes (the maximum is 20) to be created
                on the table.
                Each global secondary index in the array includes the following:
                    IndexName - The name of the global secondary index. Must be unique only for this table.
                    KeySchema - Specifies the key schema for the global secondary index.
                    Projection - Specifies attributes that are copied (projected) from the table into the index. Each
                        attribute specification is composed of:
                            ProjectionType - One of the following:
                                    KEYS_ONLY - Only the index and primary keys are projected into the index.
                                    INCLUDE - Only the specified table attributes are projected into the index.
                                    ALL - All the table attributes are projected into the index.
                            NonKeyAttributes - A list of one or more non-key attribute names that are projected into the
                                secondary index.
                ProvisionedThroughput - The provisioned throughput settings for the global secondary index, consisting
                        of read and write capacity units.
        local_secondary_indexes(list, optional): One or more local secondary indexes (the maximum is 5) to be created on
                the table. Each index is scoped to a given partition key value.

                Each local secondary index in the array includes the following:
                    IndexName - The name of the local secondary index. Must be unique only for this table.
                    KeySchema - Specifies the key schema for the local secondary index. The key schema must begin with
                        the same partition key as the table.
                    Projection - Specifies attributes that are copied (projected) from the table into the index. Each
                        attribute specification is composed of:
                            ProjectionType - One of the following:
                                KEYS_ONLY - Only the index and primary keys are projected into the index.
                                INCLUDE - Only the specified table attributes are projected into the index. The list of
                                    projected attributes is in NonKeyAttributes.
                                ALL - All the table attributes are projected into the index.
                            NonKeyAttributes - A list of one or more non-key attribute names that are projected into
                                the secondary index.
        provisioned_throughput(dict, optional): Represents the provisioned throughput settings for a specified table or
            index. If user sets BillingMode as PROVISIONED, you must specify this property. If user sets BillingMode as
            PAY_PER_REQUEST, you cannot specify this property.
        stream_specification(dict, optional): The settings for DynamoDB Streams on the table. These settings consist of:
            StreamEnabled - Indicates whether DynamoDB Streams is to be enabled (true) or disabled (false).
            StreamViewType - When an item in the table is modified, StreamViewType determines what information is
                written to the table's stream. Valid values for StreamViewType are:
                    KEYS_ONLY - Only the key attributes of the modified item are written to the stream.
                    NEW_IMAGE - The entire item, as it appears after it was modified, is written to the stream.
                    OLD_IMAGE - The entire item, as it appeared before it was modified, is written to the stream.
                    NEW_AND_OLD_IMAGES - Both the new and the old item images of the item are written to the stream.
        sse_specification(dict, optional): Represents the settings used to enable server-side encryption.
        tags(List, optional): The tags to assign to the route_table. Defaults to None.
            * Key (string) -- The key of the tag. Tag keys are case-sensitive and accept a maximum of 127 Unicode
                    characters. May not begin with aws: .
            * Value (string) -- The value of the tag. Tag values are case-sensitive and accept a maximum of 255 Unicode
                characters.
        timeout(Dict, optional): Timeout configuration for create/update/deletion of AWS DynamoDB Table.
            * create (Dict) -- Timeout configuration for creating DynamoDB Table
                * delay -- The amount of time in seconds to wait between attempts.
                * max_attempts -- Customized timeout configuration containing delay and max attempts.
            * update (string) -- Timeout configuration for updating DynamoDB Table
                * delay -- The amount of time in seconds to wait between attempts.
                * max_attempts -- Customized timeout configuration containing delay and max attempts.

    Request Syntax:
        [dynamodb_table-name]:
          aws.dynamodb.table.present:
          - name: 'string'
          - attribute_definitions: 'list'
          - key_schema: 'list'
          - resource_id: 'string'
          - billing_mode: 'string'
          - sse_specification: 'dict'
          - stream_specification: 'dict'
          - provisioned_throughput: 'dict'
          - local_secondary_indexes: 'list'
          - global_secondary_indexes: 'list'
          - tags:
            - Key: 'string'
              Value: 'string'
          - table_class: 'str'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            test_dyanmodb_table:
              aws.dynamodb.table.present:
                - name: test_dyanmodb_table
                - resource_id: dyanmodb_table-1e8965drf2c56902g
                - attribute_definitions: [
                            {"AttributeName": "Artist", "AttributeType": "S", },
                            { "AttributeName": "SongTitle", "AttributeType": "S",},
                        ]
                - key_schema = [
                            { "AttributeName": "Artist", "KeyType": "HASH", },
                            { "AttributeName": "SongTitle", "KeyType": "RANGE", },
                        ]
                - provisioned_throughput = { "ReadCapacityUnits": 123, "WriteCapacityUnits": 123, }
                - tags:
                  - Key: Name
                    Value: test-nat-gateway
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    is_table_updated = False
    existing_tags = None
    before = None
    plan_state = {}
    are_tags_updated = False
    table_arn = None
    if resource_id:
        before = await hub.exec.boto3.client.dynamodb.describe_table(
            ctx, TableName=resource_id
        )
    if before and before["result"]:
        result["comment"] = (f"'{name}' is already present",)
        table_resource = before["ret"].get("Table")
        table_arn = table_resource.get("TableArn")
        tags_ret = await hub.exec.boto3.client.dynamodb.list_tags_of_resource(
            ctx, ResourceArn=table_arn
        )
        result["result"] = tags_ret.get("result")
        if result["result"]:
            existing_tags = tags_ret.get("ret").get("Tags")
        else:
            result["comment"] = result["comment"] + (tags_ret["comment"],)
            return result
        result[
            "old_state"
        ] = hub.tool.aws.dynamodb.conversion_utils.convert_raw_dynamodb_table_to_present(
            raw_resource=before["ret"].get("Table"),
            idem_resource_name=name,
            tags=existing_tags,
        )
        result["old_state"]["tags"] = existing_tags
        plan_state = copy.deepcopy(result["old_state"])
        update_ret = (
            await hub.exec.aws.dynamodb.table.compare_inputs_and_update_dynamodb_table(
                ctx=ctx,
                plan_state=plan_state,
                attribute_definitions=attribute_definitions,
                name=name,
                billing_mode=billing_mode,
                provisioned_throughput=provisioned_throughput,
                global_secondary_indexes=global_secondary_indexes,
                sse_specification=sse_specification,
                stream_specification=stream_specification,
                replica_updates=replica_updates,
                table_class=table_class,
                timeout=timeout,
            )
        )
        result["result"] = result["result"] and update_ret["result"]
        result["comment"] = result["comment"] + update_ret["comment"]
        if result["result"]:
            is_table_updated = True
        # get existing tags
        if tags is not None:
            # Update tags
            update_tags_result = await hub.exec.aws.dynamodb.tag.update_tags(
                ctx=ctx,
                resource_arn=table_arn,
                old_tags=existing_tags,
                new_tags=tags,
            )
            result["comment"] = result["comment"] + update_tags_result["comment"]
            result["result"] = result["result"] and update_tags_result["result"]

            if not result["result"]:
                return result
            are_tags_updated = result["result"]

            if ctx.get("test", False) and (update_tags_result["ret"] is not None):
                plan_state["tags"] = update_tags_result["ret"].get("tags")
    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "attribute_definitions": attribute_definitions,
                    "key_schema": key_schema,
                    "name": name,
                    "stream_specification": stream_specification,
                    "billing_mode": billing_mode,
                    "table_class": table_class,
                    "local_secondary_indexes": local_secondary_indexes,
                    "global_secondary_indexes": global_secondary_indexes,
                    "sse_specification": sse_specification,
                    "provisioned_throughput": provisioned_throughput,
                    "replica_updates": replica_updates,
                    "tags": tags,
                    "resource_id": name,
                },
            )
            result["comment"] = (f"Would create aws.dynamodb.table '{name}'",)
            return result
        try:
            ret = await hub.exec.boto3.client.dynamodb.create_table(
                ctx,
                AttributeDefinitions=attribute_definitions,
                KeySchema=key_schema,
                TableName=name,
                BillingMode=billing_mode,
                GlobalSecondaryIndexes=global_secondary_indexes,
                LocalSecondaryIndexes=local_secondary_indexes,
                ProvisionedThroughput=provisioned_throughput,
                SSESpecification=sse_specification,
                StreamSpecification=stream_specification,
                Tags=tags if tags is not None else None,
                TableClass=table_class,
            )
            result["result"] = ret["result"]
            if not ret["result"]:
                result["comment"] = ret["comment"]
                return result
            waiter_config = hub.tool.aws.waiter_utils.create_waiter_config(
                default_delay=15,
                default_max_attempts=40,
                timeout_config=timeout.get("create") if timeout else None,
            )
            hub.log.debug(f"Waiting on creating aws.dynamodb.table '{name}'")
            try:
                await hub.tool.boto3.client.wait(
                    ctx,
                    "dynamodb",
                    "table_exists",
                    TableName=name,
                    WaiterConfig=waiter_config,
                )
            except Exception as e:
                result["comment"] = result["comment"] + (str(e),)
                result["result"] = False
            resource_id = ret["ret"]["TableDescription"]["TableName"]
            table_arn = ret["ret"].get("TableDescription").get("TableArn")
            existing_tags = tags
            result["comment"] = result["comment"] + (
                f"Created aws.dynamodb.table '{resource_id}'",
            )
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)
            result["result"] = False

    if ctx.get("test", False):
        result["new_state"] = plan_state
    elif not (before and before["result"]) or is_table_updated:
        after = await hub.exec.boto3.client.dynamodb.describe_table(
            ctx, TableName=resource_id
        )
        final_updated_tags = existing_tags
        if are_tags_updated:
            tags = await hub.exec.boto3.client.dynamodb.list_tags_of_resource(
                ctx, ResourceArn=table_arn
            )
            if tags.get("result") and tags.get("ret"):
                final_updated_tags = tags.get("ret").get("Tags")
        result[
            "new_state"
        ] = hub.tool.aws.dynamodb.conversion_utils.convert_raw_dynamodb_table_to_present(
            raw_resource=after["ret"].get("Table"),
            idem_resource_name=name,
            tags=final_updated_tags,
        )
    else:
        result["new_state"] = copy.deepcopy(result["old_state"])
    return result


async def absent(
    hub,
    ctx,
    name: str,
    resource_id: str,
    timeout: Dict = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    The DeleteTable operation deletes a table and all of its items. After a DeleteTable request, the specified table
    is in the DELETING state until DynamoDB completes the deletion. If the table is in the ACTIVE state, you can
    delete it. If a table is in CREATING or UPDATING states, then DynamoDB returns a ResourceInUseException. If the
    specified table does not exist, DynamoDB returns a ResourceNotFoundException. If table is already in the
    DELETING state, no error is returned.   DynamoDB might continue to accept data read and write operations, such
    as GetItem and PutItem, on a table in the DELETING state until the table deletion is complete.  When you delete
    a table, any indexes on that table are also deleted. If you have DynamoDB Streams enabled on the table, then the
    corresponding stream on that table goes into the DISABLED state, and the stream is automatically deleted after
    24 hours. Use the DescribeTable action to check the status of the table.

    Args:
        resource_id: A name of the DynamoDB table.
        name(Text): The Idem name of the dynamoDB table.
        timeout(Dict, optional): Timeout configuration for create/update/deletion of AWS DynamoDB Table.
            * delete (string) -- Timeout configuration for deletion of a DynamoDB Table
                * delay -- The amount of time in seconds to wait between attempts.
                * max_attempts -- Customized timeout configuration containing delay and max attempts.

    Request Syntax:
        [dynamodb_table-name]:
          aws.dynamodb.table.absent:
          - name: 'string'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            test_dyanmoDB-table:
              aws.dynamodb.table.present:
                - name: test_dynamoDB-table
                - resource_id: dyanmoDB_table-1e8965drf2c56902g:
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = await hub.exec.boto3.client.dynamodb.describe_table(
        ctx, TableName=resource_id
    )
    if not before:
        result["comment"] = (f"'{name}' already absent",)
    else:
        tags = await hub.exec.boto3.client.dynamodb.list_tags_of_resource(
            ctx, ResourceArn=before["ret"].get("Table").get("TableArn")
        )
        if tags.get("result"):
            existing_tags = tags.get("ret").get("Tags")
        else:
            result["comment"] = result["comment"] + tags["comment"]
            result["result"] = False
            return result

        result[
            "old_state"
        ] = hub.tool.aws.dynamodb.conversion_utils.convert_raw_dynamodb_table_to_present(
            raw_resource=before["ret"].get("Table"),
            idem_resource_name=name,
            tags=existing_tags,
        )
        if ctx.get("test", False):
            result["comment"] = (f"Would delete aws.dynamodb.table {name}",)
            return result
        else:
            try:
                ret = await hub.exec.boto3.client.dynamodb.delete_table(
                    ctx, TableName=resource_id
                )
                result["result"] = ret["result"]
                if not result["result"]:
                    result["comment"] = ret["comment"]
                    return result
                waiter_config = hub.tool.aws.waiter_utils.create_waiter_config(
                    default_delay=15,
                    default_max_attempts=40,
                    timeout_config=timeout.get("delete") if timeout else None,
                )
                hub.log.debug(f"Waiting on deleting aws.dynamodb.table '{name}'")
                try:
                    await hub.tool.boto3.client.wait(
                        ctx,
                        "dynamodb",
                        "table_not_exists",
                        TableName=name,
                        WaiterConfig=waiter_config,
                    )
                except Exception as e:
                    result["comment"] = result["comment"] + (str(e),)
                    result["result"] = False
                result["comment"] = result["comment"] + (
                    f"Deleted aws.dynamodb.table '{name}'",
                )
                after = await hub.exec.boto3.client.dynamodb.describe_table(
                    ctx, TableName=resource_id
                )
                if after["result"] and after["ret"]:
                    result[
                        "new_state"
                    ] = hub.tool.aws.dynamodb.conversion_utils.convert_raw_dynamodb_table_to_present(
                        raw_resource=after["ret"].get("Table"),
                        idem_resource_name=name,
                        tags=existing_tags,
                    )
            except hub.tool.boto3.exception.ClientError as e:
                result["comment"] = result["comment"] + (
                    f"{e.__class__.__name__}: {e}",
                )
                result["result"] = False
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    **Autogenerated function**

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function


    Returns an array of table names associated with the current account and endpoint. The output from ListTables is
    paginated, with each page returning a maximum of 100 table names.


    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws.dynamodb.table
    """

    result = {}

    ret = await hub.exec.boto3.client.dynamodb.list_tables(ctx)
    if not ret["result"]:
        hub.log.debug(f"Could not describe table {ret['comment']}")
        return result
    if ret and ret["ret"] and ret["ret"].get("TableNames"):
        for table_name in ret["ret"].get("TableNames"):
            resource = await hub.exec.boto3.client.dynamodb.describe_table(
                ctx, TableName=table_name
            )
            if (
                resource.get("result")
                and resource.get("ret") is not None
                and resource.get("ret").get("Table")
            ):
                table_resource = resource.get("ret").get("Table")
            else:
                continue
            tags = await hub.exec.boto3.client.dynamodb.list_tags_of_resource(
                ctx, ResourceArn=table_resource.get("TableArn")
            )
            result[table_name] = {
                "aws.dynamodb.table.present": [
                    hub.tool.aws.dynamodb.conversion_utils.convert_raw_dynamodb_table_to_present(
                        raw_resource=table_resource,
                        idem_resource_name=table_name,
                        tags=tags.get("ret").get("Tags")
                        if tags.get("result")
                        else None,
                    )
                ]
            }
    return result



"""
Autogenerated state module using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

"""
from collections import OrderedDict
from typing import Any
from typing import Dict

import dict_tools.differ as differ

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    acl: str = None,
    create_bucket_configuration: dict = None,
    grant_full_control: str = None,
    grant_read: str = None,
    grant_read_acp: str = None,
    grant_write: str = None,
    grant_write_acp: str = None,
    object_lock_enabled_for_bucket: bool = None,
    object_ownership: str = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**
        ACL (string) -- The canned ACL to apply to the bucket.

        name (string) --
        [REQUIRED]

        The name of the bucket to create.

        CreateBucketConfiguration (dict) --
            The configuration information for the bucket.

        LocationConstraint (string) --
            Specifies the Region where the bucket will be created.
            If you don't specify a Region, the bucket is created
            in the US East (N. Virginia) Region (us-east-1).

        GrantFullControl (string) --
            Allows grantee the read, write, read ACP, and write ACP permissions on the bucket.

        GrantRead (string) --
            Allows grantee to list the objects in the bucket.

        GrantReadACP (string) --
            Allows grantee to read the bucket ACL.

        GrantWrite (string) --
            Allows grantee to create new objects in the bucket.

        For the bucket and object owners of existing objects, also allows deletions and overwrites of those objects.

        GrantWriteACP (string) --
            Allows grantee to write the ACL for the applicable bucket.

        ObjectLockEnabledForBucket (boolean) --
            Specifies whether you want S3 Object Lock to be enabled for the new bucket.

        ObjectOwnership (string) --
            The container element for object ownership for a bucket's ownership controls.

        BucketOwnerPreferred -
            Objects uploaded to the bucket change ownership to the bucket owner if
            the objects are uploaded with the bucket-owner-full-control canned ACL.

        ObjectWriter -
            The uploading account will own the object if the object is uploaded
            with the bucket-owner-full-control canned ACL.

        BucketOwnerEnforced -
            Access control lists (ACLs) are disabled and no longer affect permissions.
            The bucket owner automatically owns and has full control over every object
            in the bucket. The bucket only accepts PUT requests that don't specify an
            ACL or bucket owner full control ACLs, such as the bucket-owner-full-control
            canned ACL or an equivalent form of this ACL expressed in the XML format.


    Request Syntax:
        resource:
          aws.s3.bucket.present:
            - ACL: 'private'|'public-read'|'public-read-write'|'authenticated-read'
            - Bucket: string
            - CreateBucketConfiguration: {
                'af-south-1'|'ap-east-1'|'ap-northeast-1'|'ap-northeast-2'|
                'ap-northeast-3'|'ap-south-1'|'ap-southeast-1'|'ap-southeast-2'|'ca-central-1'|
                'cn-north-1'|'cn-northwest-1'|'EU'|'eu-central-1'|'eu-north-1'|'eu-south-1'|
                'eu-west-1'|'eu-west-2'|'eu-west-3'|'me-south-1'|'sa-east-1'|'us-east-2'|
                'us-gov-east-1'|'us-gov-west-1'|'us-west-1'|'us-west-2'}
            - GrantFullControl: string
            - GrantRead: string
            - GrantReadACP: string
            - GrantWrite: string
            - GrantWriteACP: string
            - ObjectLockEnabledForBucket: True|False
            - ObjectOwnership: BucketOwnerPreferred'|'ObjectWriter'|'BucketOwnerEnforced'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            test_bucket-bb7bb32e9533:
              aws.s3.bucket.present:
                - name: test_bucket-bb7bb32e9533
                - acl: 'private'
                - create_bucket_configuration: 'us-west-1'
                - object_lock_enabled_for_bucket: False
                - object_ownership: 'BucketOwnerEnforced'


    """
    result = dict(comment="", old_state=None, new_state=None, name=name, result=True)
    resource = hub.tool.boto3.resource.create(ctx, "s3", "Bucket", name)
    before = await hub.tool.boto3.resource.describe(resource)
    result["new_state"] = before

    if ctx.get("test", False):
        if before:
            result["comment"] = f"Would update aws.s3.bucket {name}"
            result["result"] = True
        else:
            result["comment"] = f"Would create aws.s3.bucket {name}"
            result["result"] = True
        return result

    if not before:
        try:
            # Build out S3 bucket creation here
            update_ret = await hub.exec.boto3.client.s3.create_bucket(
                ctx,
                **{
                    "ACL": acl,
                    "Bucket": name,
                    "CreateBucketConfiguration": create_bucket_configuration,
                    "GrantFullControl": grant_full_control,
                    "GrantRead": grant_read,
                    "GrantReadACP": grant_read_acp,
                    "GrantWrite": grant_write,
                    "GrantWriteACP": grant_write_acp,
                    "ObjectLockEnabledForBucket": object_lock_enabled_for_bucket,
                    "ObjectOwnership": object_ownership,
                },
            )
            if not update_ret["result"]:
                result["comment"] = update_ret["comment"]
                result["result"] = False
                return result
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = f"{e.__class__.__name__}: {e}"
            result["result"] = False
            return result
        result["comment"] = f"'Created {name}'"
    else:
        result["comment"] = f"Resource {name} already exists."

    if not before:
        try:
            after = await hub.tool.boto3.resource.describe(resource)
        except Exception as e:
            result["comment"] = str(e)
            result["result"] = False
        else:
            result["old_state"] = before
            result["new_state"] = after
    return result


async def absent(hub, ctx, name: str) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Deletes the specified s3 bucket.

    Args:
        name(Text): The name of the s3 bucket.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            bucket-5435423646-456464:
              aws.s3.bucket.absent:
                - name: value
    """

    result = dict(comment="", old_state=None, new_state=None, name=name, result=True)
    resource = hub.tool.boto3.resource.create(ctx, "s3", "Bucket", name)
    before = await hub.tool.boto3.resource.describe(resource)

    if not before:
        result["comment"] = f"'{name}' already absent"
    elif ctx.get("test", False):
        result["comment"] = f"Would delete aws.s3.bucket {name}"
        return result
    else:
        try:
            ret = await hub.exec.boto3.client.s3.delete_bucket(ctx, **{"Bucket": name})
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                result["result"] = False
                return result
            result["comment"] = f"Deleted '{name}'"
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = f"{e.__class__.__name__}: {e}"
        result["old_state"] = before

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    """
    Obtain S3 buckets under the given context for any user.
    :param hub:
    :param ctx:
    :return:
    """
    result = {}
    ret = await hub.exec.boto3.client.s3.list_buckets(ctx)

    if not ret["result"]:
        hub.log.debug(f"Could not describe S3 buckets {ret['comment']}")
        return {}

    for bucket in ret["ret"]["Buckets"]:
        resource_id = bucket.get("Name")
        describe_parameters = OrderedDict(
            {
                "Name": "name",
                "CreationDate": "creation_date",
            }
        )
        resource_translated = []
        for parameter_old_key, parameter_new_key in describe_parameters.items():
            if bucket.get(parameter_old_key) is not None:
                resource_translated.append(
                    {parameter_new_key: bucket.get(parameter_old_key)}
                )
        result[resource_id] = {"aws.s3.bucket.present": resource_translated}
    return result



"""
Autogenerated using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

hub.exec.boto3.client.elasticache.create_cache_subnet_group
hub.exec.boto3.client.elasticache.delete_cache_subnet_group
hub.exec.boto3.client.elasticache.describe_cache_subnet_groups
hub.exec.boto3.client.elasticache.modify_cache_subnet_group
"""
import copy
from typing import Any
from typing import Dict
from typing import List


__contracts__ = ["resource"]
TREQ = {
    "absent": {
        "require": [
            "aws.ec2.subnet.absent",
        ],
    },
    "present": {
        "require": [
            "aws.ec2.subnet.present",
        ],
    },
}


async def present(
    hub,
    ctx,
    name: str,
    subnet_ids: List,
    cache_subnet_group_description: str,
    resource_id: str = None,
    tags: List = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Creates a new cache subnet group. Use cache subnet group parameter only when you are creating a cluster in an Amazon Virtual
    Private Cloud (Amazon VPC).

    Args:
        name(Text): An Idem name of the resource.
        cache_subnet_group_description(Text): Description of the cache subnet group
        subnet_ids(list): A list of VPC subnet IDs for the cache subnet group.
        resource_id(Text, optional): AWS Elasticache subnet group id to identify the resource
        tags(List, optional): A list of tags to be added to this resource.
            Key (string) - The key for the tag. May not be null.
            Value (string) - The tag's value. May be null.

    Request Syntax:
        [cache_subnet_group_name]:
          aws.elasticache.cache_subnet_group.present:
          - name: 'string'
          - cache_subnet_group_description: 'string'
          - subnet_ids: 'List'
          - tags:
            - Key: 'string'
              Value: 'string'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            cache_subnet_group_name:
              aws.elasticache.cache_subnet_group.present:
                - cache_subnet_group_description: "testing demo"
                - subnet_ids:
                    - subnet-239e4b44
                - tags:
                    - Key: Name
                      Value: test-elasticache
    """

    result = dict(comment="", name=name, old_state=None, new_state=None, result=True)
    before = None
    resource_updated = False
    if resource_id:
        before = await hub.exec.boto3.client.elasticache.describe_cache_subnet_groups(
            ctx, CacheSubnetGroupName=resource_id
        )
    try:
        if before and before["result"] and before["ret"]["CacheSubnetGroups"]:
            result[
                "old_state"
            ] = await hub.tool.aws.elasticache.conversion_utils.convert_raw_elasticache_subnet_to_present(
                ctx,
                raw_resource=before["ret"]["CacheSubnetGroups"][0],
                idem_resource_name=name,
            )
            plan_state = copy.deepcopy(result["old_state"])
            if cache_subnet_group_description != result["old_state"].get(
                "cache_subnet_group_description"
            ) or subnet_ids != result["old_state"].get(
                "cache_subnet_group_description"
            ):
                if not ctx.get("test", False):
                    update_ret = await hub.exec.boto3.client.elasticache.modify_cache_subnet_group(
                        ctx=ctx,
                        CacheSubnetGroupName=name,
                        CacheSubnetGroupDescription=cache_subnet_group_description,
                        SubnetIds=subnet_ids,
                    )
                    result["comment"] = update_ret["comment"]
                    result["result"] = update_ret["result"]
                    resource_updated = resource_updated or bool(update_ret["ret"])
                else:
                    if cache_subnet_group_description is not None:
                        plan_state[
                            "cache_subnet_group_description"
                        ] = cache_subnet_group_description
                    if subnet_ids is not None:
                        plan_state["subnet_ids"] = subnet_ids
            resource_arn = before["ret"]["CacheSubnetGroups"][0]["ARN"]
            if tags is not None and result["old_state"].get("tags") != tags:
                # Update tags
                update_ret = await hub.exec.aws.elasticache.tag.update_tags(
                    ctx=ctx,
                    resource_arn=resource_arn,
                    old_tags=result["old_state"].get("tags"),
                    new_tags=tags,
                )
                result["comment"] = f"{result['comment']} . {update_ret['comment']}"
                result["result"] = result["result"] and update_ret["result"]
                resource_updated = resource_updated or bool(update_ret["ret"])
                if ctx.get("test", False) and update_ret["ret"] is not None:
                    plan_state["tags"] = update_ret["ret"].get("tags")
        else:
            if ctx.get("test", False):
                result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                    enforced_state={},
                    desired_state={
                        "name": name,
                        "cache_subnet_group_description:": cache_subnet_group_description,
                        "subnet_ids": subnet_ids,
                        "tags": tags,
                    },
                )
                result[
                    "comment"
                ] = f"Would create aws.elasticache.cache_subnet_group {name}"
                return result
            ret = await hub.exec.boto3.client.elasticache.create_cache_subnet_group(
                ctx,
                CacheSubnetGroupName=name,
                CacheSubnetGroupDescription=cache_subnet_group_description,
                SubnetIds=subnet_ids,
                Tags=tags,
            )
            resource_id = name
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                return result
            result["comment"] = f"Created '{name}'"
    except hub.tool.boto3.exception.ClientError as e:
        result["result"] = False
        result["comment"] = f"{e.__class__.__name__}: {e}"
    try:
        if ctx.get("test", False):
            result["new_state"] = plan_state
        elif (not before) or resource_updated:
            after = (
                await hub.exec.boto3.client.elasticache.describe_cache_subnet_groups(
                    ctx, CacheSubnetGroupName=resource_id
                )
            )
            result[
                "new_state"
            ] = await hub.tool.aws.elasticache.conversion_utils.convert_raw_elasticache_subnet_to_present(
                ctx,
                raw_resource=after["ret"]["CacheSubnetGroups"][0],
                idem_resource_name=name,
            )
        else:
            result["new_state"] = copy.deepcopy(result["old_state"])
    except Exception as e:
        result["comment"] = str(e)
        result["result"] = False
    return result


async def absent(hub, ctx, name: str, resource_id: str) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Deletes a cache subnet group.  You cannot delete a default cache subnet group or one that is associated with any
    clusters.

    Args:
        name(Text): An Idem name of the resource.
        resource_id(Text): The AWS ID of the VPC

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            resource_is_absent:
              aws.elasticache.cache_subnet_group.absent:
                - name: value
    """

    result = dict(comment="", name=name, old_state=None, new_state=None, result=True)

    before = await hub.exec.boto3.client.elasticache.describe_cache_subnet_groups(
        ctx, CacheSubnetGroupName=resource_id
    )
    if not before and not before["result"]:
        result["comment"] = f"'{name}' already absent"
        return result
    else:
        result[
            "old_state"
        ] = await hub.tool.aws.elasticache.conversion_utils.convert_raw_elasticache_subnet_to_present(
            ctx,
            raw_resource=before["ret"]["CacheSubnetGroups"][0],
            idem_resource_name=name,
        )
        if ctx.get("test", False):
            result[
                "comment"
            ] = f"Would delete aws.elasticache.cache_subnet_group {name}"
            return result
        try:
            ret = await hub.exec.boto3.client.elasticache.delete_cache_subnet_group(
                ctx, CacheSubnetGroupName=resource_id
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                return result
            result["comment"] = f"Deleted '{name}'"
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = f"{e.__class__.__name__}: {e}"
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    **Autogenerated function**

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function


    Returns a list of cache subnet group descriptions. If a subnet group name is specified, the list contains only
    the description of that group. This is applicable only when you have ElastiCache in VPC setup. All ElastiCache
    clusters now launch in VPC by default.


    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws.elasticache.cache_subnet_group
    """

    result = {}
    ret = await hub.exec.boto3.client.elasticache.describe_cache_subnet_groups(ctx)
    if not ret["result"]:
        hub.log.debug(f"Could not describe cache_subnet_group {ret['comment']}")
        return {}

    for resource in ret["ret"]["CacheSubnetGroups"]:
        resource_id = resource.get("CacheSubnetGroupName")
        resource_translated = await hub.tool.aws.elasticache.conversion_utils.convert_raw_elasticache_subnet_to_present(
            ctx, raw_resource=resource, idem_resource_name=resource_id
        )
        result[resource_id] = {
            "aws.elasticache.cache_subnet_group.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource_translated.items()
            ]
        }
    return result



"""
Autogenerated using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

hub.exec.boto3.client.ec2.associate_route_table
hub.exec.boto3.client.ec2.create_route_table
hub.exec.boto3.client.ec2.delete_route_table
hub.exec.boto3.client.ec2.describe_route_tables
hub.exec.boto3.client.ec2.disassociate_route_table
resource = hub.tool.boto3.resource.create(ctx, "ec2", "RouteTable", name)
hub.tool.boto3.resource.exec(resource, associate_with_subnet, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, create_route, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, create_tags, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, delete, *args, **kwargs)
"""
import copy
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]

TREQ = {
    "present": {
        "require": [
            "aws.ec2.vpc.present",
        ],
    },
}


async def present(
    hub,
    ctx,
    name: str,
    vpc_id: str,
    resource_id: str = None,
    associations: List = None,
    routes: List = None,
    propagating_vgws: List = None,
    tags: List = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Creates a route table for the specified VPC. After you create a route table, you can add routes and associate
    the table with a subnet. For more information, see Route tables in the Amazon Virtual Private Cloud User Guide.

    Args:
        name(Text): An Idem name of the route table to create.
        resource_id(Text, optional): AWS Route Table ID
        vpc_id(Text): ID of the VPC.
        associations(list, optional): The associations between the route table and one or more subnets or a gateway.
            Defaults to None.
            * SubnetId (Text) -- The ID of the subnet. A subnet ID is not returned for an implicit association.
            * GatewayId (Text) -- The ID of the internet gateway or virtual private gateway.
        routes(list, optional): The routes in the route table.
            Defaults to None.
            DestinationCidrBlock (Text) -- The IPv4 CIDR block used for the destination match.
            DestinationIpv6CidrBlock (Text) -- The IPv6 CIDR block used for the destination match.
            DestinationPrefixListId (Text) -- The prefix of the Amazon Web Service.
            EgressOnlyInternetGatewayId (Text) -- The ID of the egress-only internet gateway.
            GatewayId (Text) -- The ID of a gateway attached to your VPC.
            InstanceId (Text) -- The ID of a NAT instance in your VPC.
            InstanceOwnerId (Text) -- The ID of Amazon Web Services account that owns the instance.
            NatGatewayId (Text) -- The ID of a NAT gateway.
            TransitGatewayId (Text) -- The ID of a transit gateway.
            LocalGatewayId (Text) -- The ID of the local gateway.
            CarrierGatewayId (Text) -- The ID of the carrier gateway.
            NetworkInterfaceId (Text) -- The ID of the network interface.
            VpcPeeringConnectionId (Text) -- The ID of a VPC peering connection.
            CoreNetworkArn (Text) -- The Amazon Resource Name (ARN) of the core network.
        propagating_vgws(list, optional): Any virtual private gateway (VGW) propagating routes.
            GatewayId (Text) -- The ID of the virtual private gateway.
        tags(List, optional): The tags to assign to the route_table. Defaults to None.
            * Key (string) -- The key of the tag. Tag keys are case-sensitive and accept a maximum of 127 Unicode characters. May not begin with aws: .
            * Value (string) -- The value of the tag. Tag values are case-sensitive and accept a maximum of 255 Unicode characters.

    Request Syntax:
        [route_table-resource-name]:
          aws.ec2.route_table.present:
          - vpc_id: 'string'
          - resource_id: 'string'
          - associations:
            - GatewayId: 'string'
            - SubnetId: 'string'
          - routes:
            - DestinationCidrBlock: 'string'
              GatewayId: 'string'
          - propagating_vgws:
             - 'string'
          - tags:
            - Key: 'string'
              Value: 'string'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            test_route-table:
              aws.ec2.route_table.present:
              - vpc_id: vpc-02850adfa9f6fc916
              - resource_id: route_table-3485hydfe5f6tb998
              - associations:
                - GatewayId: vgw-0334141b528f99316
                - SubnetId: subnet-0610f7c1a12a1af6a
              - routes:
                - DestinationCidrBlock: 198.31.0.0/16
                  GatewayId: local
              - propagating_vgws:
                 - vgw-0334141b528f99316
              - tags:
                - Key: Name
                  Value: route-table-association-test
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    update_successful = False
    plan_state = {}
    before = None
    if resource_id:
        before = await hub.exec.boto3.client.ec2.describe_route_tables(
            ctx, RouteTableIds=[resource_id]
        )
    if before and before["result"]:
        try:
            result[
                "old_state"
            ] = hub.tool.aws.ec2.conversion_utils.convert_raw_route_table_to_present(
                resource=before["ret"].get("RouteTables")[0],
                idem_resource_name=name,
            )
            plan_state = copy.deepcopy(result["old_state"])
            route_table_resource = before["ret"].get("RouteTables")[0]
            update_ret = await hub.exec.aws.ec2.route_table.update_associations_routes(
                ctx=ctx,
                route_table_id=route_table_resource.get("RouteTableId"),
                old_routes=route_table_resource.get("Routes"),
                new_routes=routes,
                old_associations=route_table_resource.get("Associations"),
                new_associations=associations,
                old_propagating_vgws=route_table_resource.get("PropagatingVgws"),
                new_propagating_vgws=propagating_vgws,
            )
            result["comment"] = result["comment"] + update_ret["comment"]
            update_successful = update_ret["result"]
            if not update_ret["result"]:
                result["result"] = False
            if update_ret["ret"] and ctx.get("test", False):
                if update_ret["ret"].get("routes") is not None:
                    plan_state["routes"] = update_ret["ret"].get("routes")
                if update_ret["ret"].get("associations") is not None:
                    plan_state["associations"] = update_ret["ret"].get("associations")
                if update_ret["ret"].get("vgws") is not None:
                    plan_state["propagating_vgws"] = update_ret["ret"].get("vgws")
            if update_successful and tags is not None:
                # Update tags
                update_ret = await hub.exec.aws.ec2.tag.update_tags(
                    ctx=ctx,
                    resource_id=route_table_resource.get("RouteTableId"),
                    old_tags=route_table_resource.get("Tags"),
                    new_tags=tags,
                )
                result["comment"] = result["comment"] + update_ret["comment"]
                result["result"] = update_ret["result"]
                if (
                    ctx.get("test", False)
                    and update_ret["result"]
                    and update_ret["ret"] is not None
                ):
                    plan_state["tags"] = update_ret["ret"].get("tags")
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)
            result["result"] = False
    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "vpc_id": vpc_id,
                    "resource_id": name,
                    "associations": associations,
                    "routes": routes,
                    "propagating_vgws": propagating_vgws,
                    "tags": tags,
                },
            )
            result["comment"] = (f"Would create aws.ec2.route_table {name}",)
            return result
        try:
            ret = await hub.exec.boto3.client.ec2.create_route_table(
                ctx,
                TagSpecifications=[{"ResourceType": "route-table", "Tags": tags}]
                if tags
                else None,
                VpcId=vpc_id,
            )
            result["result"] = ret["ret"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                result["result"] = False
                return result
            resource_id = ret["ret"]["RouteTable"]["RouteTableId"]
            result["comment"] = result["comment"] + (f"Created '{resource_id}'",)

            # Associate routes, associations and virtual gateways if provided
            update_ret = await hub.exec.aws.ec2.route_table.update_associations_routes(
                ctx=ctx,
                route_table_id=resource_id,
                old_routes=[],
                new_routes=routes,
                old_associations=[],
                new_associations=associations,
                old_propagating_vgws=[],
                new_propagating_vgws=propagating_vgws,
            )
            result["comment"] = result["comment"] + update_ret["comment"]
            result["result"] = result["result"] and update_ret["result"]
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)
            result["result"] = False
    try:
        if ctx.get("test", False):
            result["new_state"] = plan_state
        elif update_successful or not (before and before["result"]):
            after = await hub.exec.boto3.client.ec2.describe_route_tables(
                ctx, RouteTableIds=[resource_id]
            )
            result[
                "new_state"
            ] = hub.tool.aws.ec2.conversion_utils.convert_raw_route_table_to_present(
                resource=after["ret"].get("RouteTables")[0],
                idem_resource_name=name,
            )
        else:
            result["new_state"] = copy.deepcopy(result["old_state"])
    except Exception as e:
        result["comment"] = result["comment"] + (str(e),)
        result["result"] = False
    return result


async def absent(
    hub,
    ctx,
    name: str,
    resource_id: str,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Deletes the specified route table. You must disassociate the route table from any subnets before you can delete
    it. You can't delete the main route table.

    Args:
        resource_id: (Text): The AWS ID of the EC2 Route Table
        name(Text): A name or ID to identify the resource.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            resource_is_absent:
              aws.ec2.route_table.absent:
                - name: value
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = await hub.exec.boto3.client.ec2.describe_route_tables(
        ctx, RouteTableIds=[resource_id]
    )
    if not before["result"]:
        result["comment"] = (f"'{resource_id}' already absent",)
    else:
        result[
            "old_state"
        ] = hub.tool.aws.ec2.conversion_utils.convert_raw_route_table_to_present(
            resource=before["ret"]["RouteTables"][0],
            idem_resource_name=name,
        )
        if ctx.get("test", False):
            result["comment"] = (f"Would delete aws.ec2.route_table '{resource_id}'",)
            return result
        else:
            try:
                # we have to disassociate all subnets and gateways before we can delete route table
                if before["ret"] and before["ret"]["RouteTables"]:
                    old_route_table = before["ret"]["RouteTables"][0]
                    if old_route_table.get("Associations"):
                        for association in old_route_table.get("Associations"):
                            if association.get("Main") is not True:
                                ret = await hub.exec.boto3.client.ec2.disassociate_route_table(
                                    ctx,
                                    AssociationId=association.get(
                                        "RouteTableAssociationId"
                                    ),
                                )
                                if not ret.get("result"):
                                    result["comment"] = (
                                        result["comment"] + ret["comment"]
                                    )
                                    result["result"] = False
                                    return result
                ret = await hub.exec.boto3.client.ec2.delete_route_table(
                    ctx, RouteTableId=resource_id
                )
                result["result"] = ret["result"]
                if not result["result"]:
                    result["comment"] = result["comment"] + ret["comment"]
                    result["result"] = False
                    return result
                result["comment"] = result["comment"] + (f"Deleted '{resource_id}'",)
            except hub.tool.boto3.exception.ClientError as e:
                result["comment"] = result["comment"] + (
                    f"{e.__class__.__name__}: {e}",
                )
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    **Autogenerated function**

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function


    Describes one or more of your route tables. Each subnet in your VPC must be associated with a route table. If a
    subnet is not explicitly associated with any route table, it is implicitly associated with the main route table.
    This command does not return the subnet ID for implicit associations. For more information, see Route tables in
    the Amazon Virtual Private Cloud User Guide.


    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws.ec2.route_table
    """

    result = {}
    ret = await hub.exec.boto3.client.ec2.describe_route_tables(ctx)
    if not ret["result"]:
        hub.log.debug(f"Could not describe route_table {ret['comment']}")
        return result
    for route_table in ret["ret"]["RouteTables"]:
        route_table_id = route_table.get("RouteTableId")
        resource_converted = (
            hub.tool.aws.ec2.conversion_utils.convert_raw_route_table_to_present(
                resource=route_table,
                idem_resource_name=route_table_id,
            )
        )
        result[route_table_id] = {"aws.ec2.route_table.present": [resource_converted]}
    return result



import copy
from typing import Any
from typing import Dict


async def update_tags(
    hub,
    ctx,
    resource_name: str,
    old_tags: Dict[str, Any],
    new_tags: Dict[str, Any],
):
    """

    Args:
        hub:
        ctx:
        resource_name: aws cloudwatch resource name
        old_tags: Dict of existing tags
        new_tags: Dict of new tags

    Returns:
        {"result": True|False, "comment": "A message", "ret": None}

    """

    result = dict(comment="", result=True, ret=None)
    tags_to_add = {}
    tags_to_delete = []
    tags_result = copy.deepcopy(old_tags)
    for key, value in new_tags.items():
        if (key in old_tags and old_tags.get(key) != new_tags.get(key)) or (
            key not in old_tags
        ):
            tags_to_add[key] = value

    for key in old_tags:
        if key not in new_tags:
            tags_to_delete.append(key)
    try:
        if not ctx.get("test", False) and tags_to_delete:
            delete_tag_resp = await hub.exec.boto3.client.logs.untag_log_group(
                ctx, logGroupName=resource_name, tags=tags_to_delete
            )
            if not delete_tag_resp["result"]:
                hub.log.debug(f"Could not delete tags {tags_to_delete}")
                result["comment"] = delete_tag_resp["comment"]
                result["result"] = False
                return result
            else:
                hub.log.debug(f"Deleted tags {tags_to_delete}")
        [tags_result.pop(key, None) for key in tags_to_delete]
    except hub.tool.boto3.exception.ClientError as e:
        hub.log.debug(f"Error while deleting tags {tags_to_delete}")
        result["comment"] = f"{e.__class__.__name__}: {e}"
        result["result"] = False

    try:
        if not ctx.get("test", False) and tags_to_add:
            create_tag_resp = await hub.exec.boto3.client.logs.tag_log_group(
                ctx, logGroupName=resource_name, tags=tags_to_add
            )
            if not create_tag_resp["result"]:
                hub.log.debug(f"Could not create tags {tags_to_add}")
                result["comment"] = create_tag_resp["comment"]
                result["result"] = False
                return result
            else:
                hub.log.debug(f"Created tags {tags_to_add}")
    except hub.tool.boto3.exception.ClientError as e:
        hub.log.debug(f"Error while creating tags {tags_to_add}")
        result["comment"] = f"{e.__class__.__name__}: {e}"
        result["result"] = False
    result["ret"] = {"tags": list(tags_result.values()) + list(tags_to_add)}
    result["comment"] = f"Update tags: Add [{tags_to_add}] Remove [{tags_to_delete}]"
    result["result"] = True
    return result



"""
Autogenerated state module using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

hub.exec.boto3.client.logs.describe_log_groups
hub.exec.boto3.client.logs.create_log_group
hub.exec.boto3.client.logs.delete_log_group
"""
import copy
from typing import Any
from typing import Dict


__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    kms_key_id: str = None,
    tags: dict = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Creates a log group if it does not already exist. You can create up to 5000 log groups per account.
    You must use the following guidelines when naming a log group:
        - Log group names must be unique within a region for an AWS account.
        - Log group names can be between 1 and 512 characters long.
        - Log group names consist of the following characters: a-z, A-Z, 0-9, '_' (underscore), '-' (hyphen), '/' (forward slash), and '.' (period).
    If you associate a AWS Key Management Service (AWS KMS) customer master key (CMK) with the log group, ingested data is encrypted using the CMK.
    This association is stored as long as the data encrypted with the CMK is still within Amazon CloudWatch Logs. This enables Amazon CloudWatch Logs
    to decrypt this data whenever it is requested.
    If you attempt to associate a CMK with the log group but the CMK does not exist or the CMK is disabled, you will receive an InvalidParameterException error.

    Args:
        name(Text): An Idem name of the resource.
        resource_id(Text, optional): AWS Cloudwatch log group ID
        kms_key_id(Text, optional): The Amazon Resource Name (ARN) of the CMK to use when encrypting log data.
        tags(dict, optional): The key-value pairs to use for the tags.

    Request Syntax:
        [log_group_name]:
          aws.cloudwatch.log_group.present:
          - name: 'string'
          - kms_key_id: 'string'
          - tags:
            - 'string': 'string'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            resource_is_present:
              aws.cloudwatch.log_group.present:
                - name: value
                - tags:
                     TestKey4: TestValue4
                - kms_key_id: "arn:aws:kms:us-east-1:xxxxxxxxx:key/xxxx-9500-xxxxx"
    """

    result = dict(comment="", name=name, old_state=None, new_state=None, result=True)
    before = None
    resource_updated = False
    if resource_id:
        before = await hub.exec.boto3.client.logs.describe_log_groups(
            ctx, logGroupNamePrefix=resource_id
        )
    try:
        if before and before["ret"]["logGroups"]:
            result[
                "old_state"
            ] = await hub.tool.aws.cloudwatch.conversion_utils.convert_raw_log_group_to_present(
                ctx, raw_resource=before["ret"]["logGroups"][0], idem_resource_name=name
            )
            plan_state = copy.deepcopy(result["old_state"])
            result["comment"] = f"'{name}' already present."
            if tags is not None and dict(result["old_state"].get("tags")) != tags:
                update_ret = await hub.exec.aws.cloudwatch.tag.update_tags(
                    ctx=ctx,
                    resource_name=resource_id,
                    old_tags=result["old_state"].get("tags"),
                    new_tags=tags,
                )
                result["comment"] = f"{result['comment']} . {update_ret['comment']}"
                result["result"] = update_ret["result"]
                resource_updated = resource_updated or bool(update_ret["ret"])
                if ctx.get("test", False) and update_ret["ret"] is not None:
                    plan_state["tags"] = update_ret["ret"].get("tags")
        else:
            if ctx.get("test", False):
                result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                    enforced_state={},
                    desired_state={
                        "name": name,
                        "kms_key_id": kms_key_id,
                        "tags": tags,
                    },
                )
                result["comment"] = f"Would create aws.cloudwatch.log_group {name}"
                return result
            ret = await hub.exec.boto3.client.logs.create_log_group(
                ctx, logGroupName=name, kmsKeyId=kms_key_id, tags=tags
            )
            resource_id = name
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                return result
            ret["comment"] = f"Created '{name}'"
    except hub.tool.boto3.exception.ClientError as e:
        result["result"] = False
        result["comment"] = f"{e.__class__.__name__}: {e}"
    try:
        if ctx.get("test", False):
            result["new_state"] = plan_state
        elif (not before) or resource_updated:
            after = await hub.exec.boto3.client.logs.describe_log_groups(
                ctx, logGroupNamePrefix=resource_id
            )
            result[
                "new_state"
            ] = await hub.tool.aws.cloudwatch.conversion_utils.convert_raw_log_group_to_present(
                ctx, raw_resource=after["ret"]["logGroups"][0], idem_resource_name=name
            )
        else:
            result["new_state"] = copy.deepcopy(result["old_state"])
    except Exception as e:
        result["comment"] = str(e)
        result["result"] = False
    return result


async def absent(hub, ctx, name: str, resource_id: str) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Deletes the specified cloudwatch log group.

    Args:
        name(Text): The log Group name.
        resource_id(Text): The AWS ID of the log group

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            resource_is_absent:
              aws.cloudwatch.log_group.absent:
                - name: value
    """

    result = dict(comment="", name=name, old_state=None, new_state=None, result=True)
    before = await hub.exec.boto3.client.logs.describe_log_groups(
        ctx, logGroupNamePrefix=resource_id
    )
    if not before["ret"]["logGroups"]:
        result["comment"] = f"'{name}' already absent"
        return result
    else:
        result[
            "old_state"
        ] = await hub.tool.aws.cloudwatch.conversion_utils.convert_raw_log_group_to_present(
            ctx, raw_resource=before["ret"]["logGroups"][0], idem_resource_name=name
        )
        if ctx.get("test", False):
            result["comment"] = f"Would delete aws.cloudwatch.log_group {name}"
            return result
        try:
            ret = await hub.exec.boto3.client.logs.delete_log_group(
                ctx, logGroupName=resource_id
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                return result
            result["comment"] = f"Deleted '{name}'"
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = f"{e.__class__.__name__}: {e}"
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    **Autogenerated function**

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function


    Returns a list of cloudwatch log group descriptions. If a log group name is specified, the list contains only
    the description of that group.


    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws.cloudwatch.log_group
    """
    result = {}
    ret = await hub.exec.boto3.client.logs.describe_log_groups(ctx)
    if not ret["result"]:
        hub.log.debug(f"Could not describe log groups {ret['comment']}")
        return {}

    for resource in ret["ret"]["logGroups"]:
        resource_id = resource.get("logGroupName")
        resource_translated = await hub.tool.aws.cloudwatch.conversion_utils.convert_raw_log_group_to_present(
            ctx, raw_resource=resource, idem_resource_name=resource_id
        )
        result[resource_id] = {
            "aws.cloudwatch.log_group.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource_translated.items()
            ]
        }
    return result



import copy
from typing import Any
from typing import Dict
from typing import List


async def update_tags(
    hub,
    ctx,
    resource_id: str,
    resource_type: str,
    old_tags: List[Dict[str, Any]] = None,
    new_tags: List[Dict[str, Any]] = None,
):
    """
    Update tags of AWS route53 hosted zone resources
    Args:
        hub:
        ctx:
        resource_id (Text): route53 hosted_zone resource id
        resource_type (Text): type of resource,
        old_tags (List): list of old tags
        new_tags (List): list of new tags

    Returns:
        {"result": True|False, "comment": "A message", "ret": Dict}

    """
    tags_to_add = []
    tags_to_remove = []
    if old_tags is None:
        tags_to_remove = None
        tags_to_add = new_tags
        tags_result = {}
    else:
        old_tags_map = {tag.get("Key"): tag for tag in old_tags}
        tags_result = copy.deepcopy(old_tags_map)
        if new_tags is not None:
            for tag in new_tags:
                if tag.get("Key") in old_tags_map:
                    if tag.get("Value") == old_tags_map.get(tag.get("Key")).get(
                        "Value"
                    ):
                        del old_tags_map[tag.get("Key")]
                    else:
                        tags_to_add.append(tag)
                else:
                    tags_to_add.append(tag)
            tags_to_remove = [tag.get("Key") for tag in old_tags_map.values()]
    result = dict(comment="", result=True, ret=None)
    id = resource_id.split("/")[-1]

    if (not tags_to_remove) and (not tags_to_add):
        return result
    elif not tags_to_remove:
        tags_to_remove = None
    elif not tags_to_add:
        tags_to_add = None
    try:
        change_tag_resp = await hub.exec.boto3.client.route53.change_tags_for_resource(
            ctx,
            ResourceType=resource_type,
            ResourceId=id,
            AddTags=tags_to_add,
            RemoveTagKeys=tags_to_remove,
        )
        if not change_tag_resp["result"]:
            response_message = change_tag_resp["comment"]
            hub.log.debug(
                f"Could not modify tags for {id} with error {response_message}"
            )
            result[
                "comment"
            ] = f"Could not modify tags for {id} with error {response_message}"
            result["result"] = False
            return result
        hub.log.debug(f"modified tags for {resource_id}")
    except hub.tool.boto3.exception.ClientError as e:
        hub.log.debug(f"Error while modifying tags due to {e}")
        result["comment"] = f"{e.__class__.__name__}: {e}"
        result["result"] = False
        return result

    result["ret"] = {"tags": list(tags_result.values()) + tags_to_add}
    result[
        "comment"
    ] = f"Updated tags {old_tags} successfully with {new_tags} for {resource_id}"
    return result



"""
hub.exec.boto3.client.cloudwatch.put_metric_alarm
hub.exec.boto3.client.cloudwatch.delete_alarms
hub.exec.boto3.client.cloudwatch.describe_alarms

"""
import copy
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    period: int,
    evaluation_periods: int,
    comparison_operator: str,
    resource_id: str = None,
    alarm_description: str = None,
    actions_enabled: bool = True,
    ok_actions: List = None,
    alarm_actions: List = None,
    insufficient_data_actions: List = None,
    metric_name: str = None,
    namespace: str = None,
    statistic: str = None,
    extended_statistic: str = None,
    dimensions: List = None,
    unit: str = None,
    datapoints_to_alarm: int = None,
    threshold: float = None,
    treat_missing_data: str = None,
    evaluate_low_sample_count_percentile: str = None,
    metrics: List = None,
    threshold_metric_id: str = None,
    tags: List = None,
):
    r"""

    Creates or updates an alarm and associates it with the specified metric, metric math expression, or anomaly
    detection model.Alarms based on anomaly detection models cannot have Auto Scaling actions. When this operation
    creates an alarm, the alarm state is immediately set to INSUFFICIENT_DATA . The alarm is then evaluated and its
    state is set appropriately. Any actions associated with the new state are then executed. When you update an
    existing alarm, its state is left unchanged, but the update completely overwrites the previous configuration
    of the alarm. If you are an IAM user,
    you must have Amazon EC2 permissions for some alarm operations:
        * The iam:CreateServiceLinkedRole for all alarms with EC2 actions
        * The iam:CreateServiceLinkedRole to create an alarm with Systems Manager OpsItem actions.
    The first time you create an alarm in the Amazon Web Services Management Console, the CLI, or by using the
    PutMetricAlarm API, CloudWatch creates the necessary service-linked role for you. The service-linked roles are
    called AWSServiceRoleForCloudWatchEvents and AWSServiceRoleForCloudWatchAlarms_ActionSSM.

    Args:
        name(Text): The name for the alarm. This name must be unique within the Region.
        resource_id(Text): The AWS name of the metric alarm.
        alarm_description(Text, Optional): The description for the alarm.
        actions_enabled(Boolean, Optional): Indicates whether actions should be executed during any changes to the alarm
        state.
            The default is TRUE .
        ok_actions(List, Optional): The actions to execute when this alarm transitions to an OK state from any other
        state.
            Each action is specified as an Amazon Resource Name (ARN).
            Valid Values:
                arn:aws:automate:*region* :ec2:stop | arn:aws:automate:*region* :ec2:terminate |
                arn:aws:automate:*region* :ec2:recover | arn:aws:automate:*region* :ec2:reboot |
                ``arn:aws:sns:region :account-id :sns-topic-name `` |
                ``arn:aws:autoscaling:region :account-id :scalingPolicy:policy-id
                :autoScalingGroupName/group-friendly-name :policyName/policy-friendly-name ``
            Valid Values (for use with IAM roles):
                arn:aws:swf:*region* :*account-id* :action/actions/AWS_EC2.InstanceId.Stop/1.0 |
                arn:aws:swf:*region* :*account-id* :action/actions/AWS_EC2.InstanceId.Terminate/1.0 |
                arn:aws:swf:*region* :*account-id* :action/actions/AWS_EC2.InstanceId.Reboot/1.0 |
                arn:aws:swf:*region* :*account-id* :action/actions/AWS_EC2.InstanceId.Recover/1.0
        alarm_actions(List, Optional):The actions to execute when this alarm transitions to the ALARM state from any
        other state.
            Each action is specified as an Amazon Resource Name (ARN).
            Valid Values:
                arn:aws:automate:*region* :ec2:stop |
                arn:aws:automate:*region* :ec2:terminate |
                arn:aws:automate:*region* :ec2:recover |
                arn:aws:automate:*region* :ec2:reboot |
                ``arn:aws:sns:region :account-id :sns-topic-name `` |
                 ``arn:aws:autoscaling:region :account-id :scalingPolicy:policy-id
                 :autoScalingGroupName/group-friendly-name :policyName/policy-friendly-name `` |
                ``arn:aws:ssm:region :account-id :opsitem:severity `` |
                ``arn:aws:ssm-incidents::account-id :response-plan:response-plan-name ``
           Valid Values (for use with IAM roles):
                arn:aws:swf:*region* :*account-id* :action/actions/AWS_EC2.InstanceId.Stop/1.0 |
                arn:aws:swf:*region* :*account-id* :action/actions/AWS_EC2.InstanceId.Terminate/1.0 |
                arn:aws:swf:*region* :*account-id* :action/actions/AWS_EC2.InstanceId.Reboot/1.0 |
                arn:aws:swf:*region* :*account-id* :action/actions/AWS_EC2.InstanceId.Recover/1.0

        insufficient_data_actions(List, Optional):The actions to execute when this alarm transitions to the
        INSUFFICIENT_DATA state from any other state. Each action is specified as an Amazon Resource Name (ARN).
            Valid Values:
                arn:aws:automate:*region* :ec2:stop |
                arn:aws:automate:*region* :ec2:terminate |
                arn:aws:automate:*region* :ec2:recover |
                arn:aws:automate:*region* :ec2:reboot |
                ``arn:aws:sns:region :account-id :sns-topic-name `` |
                ``arn:aws:autoscaling:region :account-id :scalingPolicy:policy-id
                :autoScalingGroupName/group-friendly-name :policyName/policy-friendly-name ``
            Valid Values (for use with IAM roles):
                >arn:aws:swf:*region* :*account-id* :action/actions/AWS_EC2.InstanceId.Stop/1.0 |
                arn:aws:swf:*region* :*account-id* :action/actions/AWS_EC2.InstanceId.Terminate/1.0 |
                arn:aws:swf:*region* :*account-id* :action/actions/AWS_EC2.InstanceId.Reboot/1.0
        metric_name(Text, Optional): The name for the metric associated with the alarm. For each PutMetricAlarm operation,
            you must specify either MetricName or a Metrics array.
        namespace(Text, Optional): The namespace for the metric associated specified in MetricName .
        statistic(Text, Optional): The statistic for the metric specified in MetricName , other than percentile. For
            percentile statistics, use ExtendedStatistic . When you call PutMetricAlarm and specify a MetricName ,
            you must specify either Statistic or ExtendedStatistic, but not both.
        extended_statistic(Text, Optional): The percentile statistic for the metric specified in MetricName . Specify a value
            between p0.0 and p100. When you call PutMetricAlarm and specify a MetricName , you must specify either
            Statistic or ExtendedStatistic, but not both.
        dimensions(List, Optional): The dimensions for the metric specified in MetricName. A dimension is a name/value pair that
            is part of the identity of a metric. You can assign up to 10 dimensions to a metric. Because dimensions are
            part of the unique identifier for a metric, whenever you add a unique name/value pair to one of your
            metrics, you are creating a new variation of that metric.
            * Name (Text) -- The name of the dimension. Dimension names must contain only ASCII characters and must
                include at least one non-whitespace character.
            * Value (Text) -- The value of the dimension. Dimension values must contain only ASCII characters and must
                include at least one non-whitespace character.
        period(Integer): The length, in seconds, used each time the metric specified in MetricName is evaluated.
            Valid values are 10, 30, and any multiple of 60.
        unit(Text, Optional): The unit of measure for the statistic. For example, the units for the Amazon EC2 NetworkIn metric are
            Bytes because NetworkIn tracks the number of bytes that an instance receives on all network interfaces.
            You can also specify a unit when you create a custom metric. Units help provide conceptual meaning to your
            data. Metric data points that specify a unit of measure, such as Percent, are aggregated separately.
        evaluation_periods(Integer): The number of periods over which data is compared to the specified threshold.
            If you are setting an alarm that requires that a number of consecutive data points be breaching to trigger
            the alarm, this value specifies that number. If you are setting an "M out of N" alarm, this value is the N.
            An alarm's total current evaluation period can be no longer than one day, so this number multiplied by
            Period cannot be more than 86,400 seconds.

        datapoints_to_alarm(Integer, Optional): The number of data points that must be breaching to trigger the alarm.
            This is used only if you are setting an "M out of N" alarm. In that case, this value is the M.
        threshold(float, optional): The value against which the specified statistic is compared.  This parameter is required for
            alarms based on static thresholds, but should not be used for alarms based on anomaly detection models.
        comparison_operator(Text): The arithmetic operation to use when comparing the specified statistic and
            threshold. The specified statistic value is used as the first operand.
            The values LessThanLowerOrGreaterThanUpperThreshold , LessThanLowerThreshold , and
            GreaterThanUpperThreshold are used only for alarms based on anomaly detection models.
        treat_missing_data(Text, Optional): Sets how this alarm is to handle missing data points. If TreatMissingData is
            omitted, the default behavior of missing is used.
            Valid Values: breaching | notBreaching | ignore | missing
        evaluate_low_sample_count_percentile(Text, Optional): Used only for alarms based on percentiles. If you specify ignore ,
            the alarm state does not change during periods with too few data points to be statistically significant.
            If you specify evaluate or omit this parameter, the alarm is always evaluated and possibly changes state
            no matter how many data points are available.
        metrics(List, Optional): An array of MetricDataQuery structures that enable you to create an alarm based on the result
            of a metric math expression. For each PutMetricAlarm operation, you must specify either MetricName or a
            Metrics array.
        threshold_metric_id(Text, Optional): If this is an alarm based on an anomaly detection model, make this value match the
            ID of the ANOMALY_DETECTION_BAND function.
        tags(List, Optional): A list of key-value pairs to associate with the alarm. You can associate as many as 50 tags with an
            alarm.
            * Key (Text) -- A string that you can use to assign a value. The combination of tag keys and values can
                help you organize and categorize your resources.
            * Value (Text) -- The value for the specified tag key.

    Request Syntax:
        [metric-alarm-name]::
            aws.cloudwatch.metric_alarm.present:
            - name: 'string'
            - alarm_description: 'string'
            - actions_enabled: True|False
            - ok_actions= ['string']
            - alarm_actions: ['string']
            - insufficient_data_actions: ['string']
            - metric_name: 'string'
            - namespace: 'string'
            - statistic: 'SampleCount'|'Average'|'Sum'|'Minimum'|'Maximum'
            - dimensions: 'list'
            - period: 'integer'
            - unit: 'string'
            - evaluation_periods: 'integer'
            - datapoints_to_alarm: 'integer'
            - threshold: 'integer'
            - comparison_operator: 'string'
            - treat_missing_data: 'string',
            - evaluate_low_sample_count_percentile: 'string',
            - metrics: 'list'
            - threshold_metric_id: 'string'
            - tags:
              - Key: 'string'
                Value: 'string'

    Returns:
         Dict[str, Any]


    Examples:

        .. code-block:: sls

            awsec2-i-0f36e2b10a7463129-LessThanOrEqualToThreshold-CPUUtilization:
              aws.cloudwatch.metric_alarm.present:
                - name: awsec2-i-0f36e2b10a7463129-LessThanOrEqualToThreshold-CPUUtilization
                - alarm_description: "stop EC2 instance if it utilizes CPU less than 5"
                - actions_enabled: True
                - alarm_actions:
                  - arn:aws:swf:*region*:*account-id*:action/actions/AWS_EC2.InstanceId.Stop/1.0
                - insufficient_data_actions: []
                - metric_name: CPUUtilization
                - namespace: AWS/EC2
                - statistic: Average
                - dimensions:
                  - Name: InstanceId
                    Value: i-0f36e2b10a7463129
                - period: 60
                - evaluation_periods: 1
                - datapoints_to_alarm: 1
                - threshold: 20
                - comparison_operator: LessThanThreshold
                - tags:
                  - Key: type
                    Value: metric_alarm

    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    resource_updated = False
    before = None
    resource_parameters = {
        "AlarmName": name,
        "AlarmDescription": alarm_description,
        "ActionsEnabled": actions_enabled,
        "OKActions": ok_actions,
        "AlarmActions": alarm_actions,
        "InsufficientDataActions": insufficient_data_actions,
        "MetricName": metric_name,
        "Namespace": namespace,
        "Statistic": statistic,
        "ExtendedStatistic": extended_statistic,
        "Dimensions": dimensions,
        "Period": period,
        "Unit": unit,
        "EvaluationPeriods": evaluation_periods,
        "DatapointsToAlarm": datapoints_to_alarm,
        "Threshold": threshold,
        "ComparisonOperator": comparison_operator,
        "TreatMissingData": treat_missing_data,
        "EvaluateLowSampleCountPercentile": evaluate_low_sample_count_percentile,
        "Metrics": metrics,
        "ThresholdMetricId": threshold_metric_id,
        "Tags": tags,
    }

    if resource_id:
        resource = hub.tool.boto3.resource.create(
            ctx, "cloudwatch", "Alarm", resource_id
        )
        before = await hub.tool.boto3.resource.describe(resource)

    if before:
        try:
            convert_ret = await hub.tool.aws.cloudwatch.conversion_utils.convert_raw_metric_alarm_to_present(
                ctx, raw_resource=before, idem_resource_name=name
            )
            result["result"] = convert_ret["result"]
            if not result["result"]:
                result["comment"] = result["comment"] + convert_ret["comment"]
            result["old_state"] = convert_ret["ret"]

            plan_state = copy.deepcopy(result["old_state"])
            # Update metric alarm
            update_ret = await hub.exec.aws.cloudwatch.metric_alarm.update_metric(
                ctx,
                alarm_name=name,
                raw_resource=before,
                resource_parameters=resource_parameters,
            )
            result["comment"] = result["comment"] + update_ret["comment"]
            result["result"] = update_ret["result"]
            resource_updated = resource_updated or bool(update_ret["ret"])
            if update_ret["ret"] and ctx.get("test", False):
                for key in [
                    "alarm_description",
                    "actions_enabled",
                    "ok_actions",
                    "alarm_actions",
                    "metric_name",
                    "namespace",
                    "statistic",
                    "dimensions",
                    "period",
                    "evaluation_periods",
                    "datapoints_to_alarm",
                    "threshold",
                    "comparison_operator",
                ]:
                    if key in update_ret["ret"]:
                        plan_state[key] = update_ret["ret"][key]
            if (tags is not None) and (
                not hub.tool.aws.state_comparison_utils.are_lists_identical(
                    tags, result["old_state"].get("tags")
                )
            ):
                # Update tags
                update_tag_ret = (
                    await hub.exec.aws.cloudwatch.metric_alarm.update_metric_tags(
                        ctx=ctx,
                        alarm_arn=before.get("AlarmArn"),
                        old_tags=result["old_state"].get("tags", []),
                        new_tags=tags,
                    )
                )
                result["result"] = result["result"] and update_tag_ret["result"]
                result["comment"] = result["comment"] + update_tag_ret["comment"]
                resource_updated = resource_updated or bool(update_tag_ret["ret"])
                if ctx.get("test", False) and update_tag_ret["ret"] is not None:
                    plan_state["tags"] = update_tag_ret["ret"].get("tags")

            if not resource_updated:
                result["comment"] = result["comment"] + (
                    f"aws.cloudwatch.metric_alarm '{name}' has no property need to be updated.",
                )
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)
            result["result"] = False
    else:
        try:
            if ctx.get("test", False):
                result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                    enforced_state={},
                    desired_state={
                        "name": name,
                        "alarm_description": alarm_description,
                        "actions_enabled": actions_enabled,
                        "ok_actions": ok_actions,
                        "alarm_actions": alarm_actions,
                        "insufficient_data_actions": insufficient_data_actions,
                        "metric_name": metric_name,
                        "namespace": namespace,
                        "statistic": statistic,
                        "extended_statistic": extended_statistic,
                        "dimensions": dimensions,
                        "period": period,
                        "unit": unit,
                        "evaluation_periods": evaluation_periods,
                        "datapoints_to_alarm": datapoints_to_alarm,
                        "threshold": threshold,
                        "comparison_operator": comparison_operator,
                        "treat_missing_data": treat_missing_data,
                        "evaluate_low_sample_count_percentile": evaluate_low_sample_count_percentile,
                        "metrics": metrics,
                        "tags": tags,
                        "threshold_metric_id": threshold_metric_id,
                    },
                )
                result["comment"] = (
                    f"Would create aws.cloudwatch.metric_alarm {name}",
                )
                return result

            # Create metric alarm
            ret = await hub.exec.boto3.client.cloudwatch.put_metric_alarm(
                ctx, **resource_parameters
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = result["comment"] + ret["comment"]
                return result
            result["comment"] = result["comment"] + (f"Created '{name}'",)
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)
            result["result"] = False

    try:
        if ctx.get("test", False):
            result["new_state"] = plan_state
        elif (not before) or resource_updated:
            resource = hub.tool.boto3.resource.create(ctx, "cloudwatch", "Alarm", name)
            after = await hub.tool.boto3.resource.describe(resource)
            convert_ret = await hub.tool.aws.cloudwatch.conversion_utils.convert_raw_metric_alarm_to_present(
                ctx, raw_resource=after, idem_resource_name=name
            )
            result["result"] = convert_ret["result"]
            if not result["result"]:
                result["comment"] = result["comment"] + convert_ret["comment"]
            result["new_state"] = convert_ret["ret"]
        else:
            result["new_state"] = copy.deepcopy(result["old_state"])
    except Exception as e:
        result["comment"] = result["comment"] + (str(e),)
        result["result"] = False

    return result


async def absent(hub, ctx, name: str, resource_id: str) -> Dict[str, Any]:
    r"""
    Deletes the specified alarms. You can delete up to 100 alarms in one operation. However, this total can include no
    more than one composite alarm. For example, you could delete 99 metric alarms and one composite alarms with one
    operation, but you can't delete two composite alarms with one operation.

    Args:
        hub:
        ctx:
        name(Text): The AWS name of the metric alarm.
        resource_id(Text): The AWS name of the metric alarm.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            [alarm name]:
               aws.cloudwatch.metric_alarm.absent:
                - resource_id: value

    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    resource = hub.tool.boto3.resource.create(ctx, "cloudwatch", "Alarm", resource_id)
    before = await hub.tool.boto3.resource.describe(resource)

    if not before:
        result["comment"] = (f"'{name}' already absent",)
    elif ctx.get("test", False):
        convert_ret = await hub.tool.aws.cloudwatch.conversion_utils.convert_raw_metric_alarm_to_present(
            ctx, raw_resource=before, idem_resource_name=name
        )
        result["result"] = convert_ret["result"]
        if not result["result"]:
            result["comment"] = result["comment"] + convert_ret["comment"]
        result["old_state"] = convert_ret["ret"]
        result["comment"] = (f"Would delete aws.cloudwatch.metric_alarm '{name}'",)
        return result
    else:
        try:
            convert_ret = await hub.tool.aws.cloudwatch.conversion_utils.convert_raw_metric_alarm_to_present(
                ctx, raw_resource=before, idem_resource_name=name
            )
            result["result"] = convert_ret["result"]
            if not result["result"]:
                result["comment"] = result["comment"] + convert_ret["comment"]
            result["old_state"] = convert_ret["ret"]
            ret = await hub.exec.boto3.client.cloudwatch.delete_alarms(
                ctx, AlarmNames=[resource_id]
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                result["result"] = False
                return result
            result["comment"] = (f"Deleted '{name}'",)

        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    Retrieves the specified alarms. You can filter the results by specifying a prefix for the alarm name, the alarm
    state, or a prefix for any action. To use this operation and return information about composite alarms, you must
    be signed on with the cloudwatch:DescribeAlarms permission that is scoped to * . You can't return information
    about composite alarms if your cloudwatch:DescribeAlarms permission has a narrower scope.

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: bash

            $ idem describe aws.cloudwatch.metric_alarm
    """
    result = {}
    ret = await hub.exec.boto3.client.cloudwatch.describe_alarms(ctx)

    if not ret["result"]:
        hub.log.debug(f"Could not describe alarm metrics {ret['comment']}")
        return {}

    for resource in ret["ret"]["MetricAlarms"]:
        alarm_name = resource["AlarmName"]
        convert_ret = await hub.tool.aws.cloudwatch.conversion_utils.convert_raw_metric_alarm_to_present(
            ctx, raw_resource=resource, idem_resource_name=alarm_name
        )
        if not convert_ret["result"]:
            hub.log.warning(
                f"Could not describe alarm metrics '{alarm_name}' with error {ret['comment']}"
            )
            continue
        translated_resource = convert_ret["ret"]
        result[translated_resource["resource_id"]] = {
            "aws.cloudwatch.metric_alarm.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in translated_resource.items()
            ]
        }

    return result



"""
Autogenerated using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

hub.exec.boto3.client.iam.create_instance_profile
hub.exec.boto3.client.iam.delete_instance_profile
hub.exec.boto3.client.iam.get_instance_profile
hub.exec.boto3.client.iam.list_instance_profiles
hub.exec.boto3.client.iam.tag_instance_profile
hub.exec.boto3.client.iam.untag_instance_profile
resource = hub.tool.boto3.resource.create(ctx, "iam", "InstanceProfile", name)
hub.tool.boto3.resource.exec(resource, add_role, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, delete, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, remove_role, *args, **kwargs)
"""
import copy
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    path: str = None,
    roles: List = None,
    tags: List = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

     Creates a new instance profile. For information about instance profiles, see Using roles for applications on
    Amazon EC2 in the IAM User Guide, and Instance profiles in the Amazon EC2 User Guide.  For information about the
    number of instance profiles you can create, see IAM object quotas in the IAM User Guide.
    Tags can be set on a newly created instance profile, or updated on an existing one.
    Roles are set and updated as a day-2 operation.

    Args:
        name(Text): Name of the resource.
        resource_id(Text, Optional): ID of the resource, once created.
        path(Text, Optional): The path to the instance profile.
        roles(Text, Optional): A list with up to one role name. Instance profile can be associated with a single role.
        tags(List, Optional): A list of tags that you want to attach to the new role. Each tag consists of a
        key name and an associated value.

    Returns:
        Dict[str, Any]

    Request Syntax:
      [iam-instance-profile-name]:
        aws.iam.instance_profile.present:
          - name: 'string'
          - resource_id: 'string'
          - path: 'string'
          - roles:
            - RoleName: 'string'
          - Tags:
            - Key: 'string'
              Value: 'string'

    Example:

        .. code-block:: sls

        inst-profile-1:
          aws.iam.instance_profile.present:
          - name: inst-profile-1
          - path: /aws-service-inst-profile/
          - roles:
            - RoleName: role-1
          - Tags:
            - Key: test-1
              Value: test-1-val

    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    resource = hub.tool.boto3.resource.create(
        ctx, "iam", "InstanceProfile", resource_id if resource_id else name
    )
    before = await hub.tool.boto3.resource.describe(resource)
    if before:
        result[
            "old_state"
        ] = hub.tool.aws.iam.conversion_utils.convert_raw_instance_profile_to_present(
            before
        )
        result["new_state"] = copy.deepcopy(result["old_state"])
        result["comment"] = (f"'{name}' already exists",)
        # Update key tags if tags are specified
        if (
            tags is not None
            and not hub.tool.aws.state_comparison_utils.are_lists_identical(
                tags, result["old_state"].get("tags", None)
            )
        ):
            hub.log.debug(f"aws.iam.instance_profile '{name}' tags update")
            update_ret = (
                await hub.exec.aws.iam.instance_profile.update_instance_profile_tags(
                    ctx=ctx,
                    instance_profile_name=resource_id if resource_id else name,
                    old_tags=before.get("tags", []),
                    new_tags=tags,
                )
            )

            result["result"] = update_ret["result"]
            result["comment"] = result["comment"] + update_ret["comment"]
            if not result["result"]:
                return result
            if ctx.get("test", False) and update_ret["ret"] is not None:
                result["new_state"]["tags"] = update_ret["ret"].get("tags")
    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "path": path,
                    "roles": roles,
                    "tags": tags,
                },
            )
            result["comment"] = (f"Would create aws.iam.instance_profile {name}",)
            return result
        try:
            ret = await hub.exec.boto3.client.iam.create_instance_profile(
                ctx,
                InstanceProfileName=name,
                Path=path,
                Tags=tags,
            )
            result["result"] = ret["result"]
            if not ret["result"]:
                result["comment"] = ret["comment"]
                return result
            result["comment"] = (f"Created '{name}'",)
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = (f"{e.__class__.__name__}: {e}",)
            result["result"] = False

    update_ret = None
    if roles is not None:
        # Instance profile can be associated only with a single role
        # Remove existing role
        curr_role = None
        new_role = None
        if before and before.get("Roles", []):
            curr_role = before.get("Roles")[0]["RoleName"]
        if roles:
            new_role = roles[0]["RoleName"]
        if curr_role != new_role:
            if curr_role is not None and not ctx.get("test", False):
                hub.log.debug(
                    f"Removing role '{curr_role}' from instance profile '{name}'"
                )
                update_ret = (
                    await hub.exec.boto3.client.iam.remove_role_from_instance_profile(
                        ctx, InstanceProfileName=name, RoleName=curr_role
                    )
                )
                if not update_ret["result"]:
                    result["result"] = False
                    result["comment"] = update_ret["comment"]
                    return result

                result["comment"] = result["comment"] + (
                    f"Role '{curr_role}' was removed from instance profile '{name}'",
                )

            if new_role is not None and not ctx.get("test", False):
                hub.log.debug(f"Adding role '{new_role}' to instance profile '{name}'")
                update_ret = (
                    await hub.exec.boto3.client.iam.add_role_to_instance_profile(
                        ctx, InstanceProfileName=name, RoleName=new_role
                    )
                )
                if not update_ret["result"]:
                    result["result"] = False
                    result["comment"] = update_ret["comment"]
                    return result

                result["comment"] = result["comment"] + (
                    f"Role '{new_role}' was added to instance profile '{name}'",
                )
            # Test context when updating roles
            if ctx.get("test", False):
                result["comment"] = result["comment"] + (
                    f"Updated role in instance profile '{name}'.",
                )
                result["new_state"]["roles"] = (
                    [{"RoleName": new_role}] if new_role is not None else []
                )
                return result

    if ctx.get("test", False):
        return result

    if before and update_ret is None:
        result["new_state"] = copy.deepcopy(result["old_state"])
    else:
        resource = hub.tool.boto3.resource.create(ctx, "iam", "InstanceProfile", name)
        after = await hub.tool.boto3.resource.describe(resource)
        result[
            "new_state"
        ] = hub.tool.aws.iam.conversion_utils.convert_raw_instance_profile_to_present(
            after
        )

    return result


async def absent(hub, ctx, name: str, resource_id: str = None) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Deletes the specified instance profile. The instance profile must not have an associated role.  Make sure that
    you do not have any Amazon EC2 instances running with the instance profile you are about to delete. Deleting a
    role or instance profile that is associated with a running instance will break any applications running on the
    instance.  For more information about instance profiles, see About instance profiles.

    Args:
        name(Text): Instance profile name
        resource_id(Text, Optional): Instance profile ID

    Returns:
        Dict[str, Any]

    Request Syntax:
      [iam-instance-profile-name]:
        aws.iam.instance_profile.present:
          - name: 'string'

    Examples:

        .. code-block:: sls

            instance-profile-test:
              aws.iam.instance_profile.absent:
                - name: instance-profile-test
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    resource = hub.tool.boto3.resource.create(
        ctx, "iam", "InstanceProfile", resource_id if resource_id else name
    )
    before = await hub.tool.boto3.resource.describe(resource)
    if not before:
        result["comment"] = (f"'{name}' already absent",)
        return result
    else:
        try:
            result[
                "old_state"
            ] = hub.tool.aws.iam.conversion_utils.convert_raw_instance_profile_to_present(
                before
            )
            if ctx.get("test", False):
                result["comment"] = (f"Would delete aws.iam.instance_profile {name}",)
                return result

            ret = await hub.exec.boto3.client.iam.delete_instance_profile(
                ctx,
                InstanceProfileName=name,
            )
            if not result["result"]:
                result["comment"] = ret["comment"]
                result["result"] = False
                return result
            result["comment"] = (f"Deleted '{name}'",)
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = (f"{e.__class__.__name__}: {e}",)

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    **Autogenerated function**

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function


    Lists the instance profiles that have the specified path prefix. If there are none, the operation returns an
    empty list. For more information about instance profiles, see About instance profiles.  IAM resource-listing
    operations return a subset of the available attributes for the resource. For example, this operation does not
    return tags, even though they are an attribute of the returned object. To view all of the information for an
    instance profile, see GetInstanceProfile.  You can paginate the results using the MaxItems and Marker
    parameters.


    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws.iam.instance_profile
    """

    result = {}
    ret = await hub.exec.boto3.client.iam.list_instance_profiles(ctx)

    if not ret["result"]:
        hub.log.debug(f"Could not describe aws.iam.instance_profile {ret['comment']}")
        return {}

    for instance_profile in ret["ret"]["InstanceProfiles"]:
        # This is required to get tags for each instance profile list_instance_profile_roles is not working
        boto2_resource = hub.tool.boto3.resource.create(
            ctx, "iam", "InstanceProfile", instance_profile["InstanceProfileName"]
        )
        resource = await hub.tool.boto3.resource.describe(boto2_resource)

        translated_resource = (
            hub.tool.aws.iam.conversion_utils.convert_raw_instance_profile_to_present(
                resource
            )
        )
        result[translated_resource["name"]] = {
            "aws.iam.instance_profile.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in translated_resource.items()
            ]
        }

    return result
